# Copyright (c) 2013 Novo Nordisk Foundation Center for Biosustainability, DTU.
# See LICENSE for details.

import sys
from unittest import TestCase

from optlang.exceptions import ContainerAlreadyContains
from optlang.interface import Model, Variable, Constraint, Objective


class TestModel(TestCase):
    def setUp(self):
        self.model = Model()
        x = Variable('x', lb=0, ub=10)
        y = Variable('y', lb=0, ub=10)
        constr = Constraint(1. * x + y, lb=3, name="constr1")
        obj = Objective(2 * x + y)
        self.model.add(x)
        self.model.add(y)
        self.model.add(constr)
        self.model.objective = obj

    def test_read_only_attributes(self):
        self.assertRaises(AttributeError, setattr, self.model, 'variables', 'Foo')
        self.assertRaises(AttributeError, setattr, self.model, 'constraints', 'Foo')
        self.assertRaises(AttributeError, setattr, self.model, 'status', 'Foo')

    def test_create_empty_model(self):
        model = Model()
        self.assertEqual(len(model.constraints), 0)
        self.assertEqual(len(model.variables), 0)
        self.assertEqual(model.objective.expression - 0, 0)

    def test_add_variable(self):
        var = Variable('z')
        self.model.add(var)
        self.assertTrue(var in self.model.variables.values())
        self.assertEqual(self.model.variables.values().count(var), 1)
        self.assertEqual(self.model.variables['z'].problem, var.problem)
        var = Variable('asdfasdflkjasdflkjasdlkjsadflkjasdflkjasdlfkjasdlfkjasdlkfjasdf', lb=-13)
        self.model.add(var)
        self.assertTrue(var in self.model.variables.values())
        self.assertEqual(self.model.variables['z'].lb, None)
        self.assertEqual(self.model.variables['z'].ub, None)
        self.assertEqual(self.model.variables['asdfasdflkjasdflkjasdlkjsadflkjasdflkjasdlfkjasdlfkjasdlkfjasdf'].lb,
                         -13)
        self.assertEqual(self.model.variables['asdfasdflkjasdflkjasdlkjsadflkjasdflkjasdlfkjasdlfkjasdlkfjasdf'].ub,
                         None)

    def test_add_variable_twice_raises(self):
        var = Variable('x')
        self.model.add(var)
        self.assertRaises(ContainerAlreadyContains, self.model.update)

    def test_remove_add_variable(self):
        var = self.model.variables[0]
        self.model.remove(var)
        self.model.add(var)
        self.model.update()

    def test_remove_add_remove_variable(self):
        var = self.model.variables[0]
        self.model.remove(var)
        self.model.add(var)
        # self.assertRaises(ContainerAlreadyContains, self.model.remove, var)

    def test_add_existing_variable(self):
        var = self.model.variables[0]
        self.model.add(var)
        self.assertRaises(Exception, self.model.update)

    def test_remove_constraint(self):
        self.model.remove('constr1')
        self.assertEqual(list(self.model.constraints), [])

    def test_add_remove_constraint(self):
        c = Constraint(self.model.variables.x + self.model.variables.y, lb=10)
        self.model.add(c)
        self.assertEqual(list(self.model.constraints), [self.model.constraints['constr1'], c])
        self.model.remove(c)
        self.model.update()
        self.assertEqual(list(self.model.constraints), [self.model.constraints['constr1']])

    def test_add_remove_collection(self):
        c = Constraint(self.model.variables.x + self.model.variables.y, lb=10)
        c2 = Constraint(3. * self.model.variables.x + self.model.variables.y, lb=10)
        self.model.add([c, c2])
        self.assertEqual(list(self.model.constraints), [self.model.constraints['constr1'], c, c2])
        self.model.remove([c, 'constr1', c2])
        self.assertEqual(list(self.model.constraints), [])

    def test_removing_objective_raises(self):
        self.assertRaises(TypeError, self.model.remove, self.model.objective)

    def test_removing_crap_raises(self):
        self.assertRaises(TypeError, self.model.remove, dict)

    def test_remove_variable_str(self):
        var = self.model.variables['y']
        self.model.remove(var.name)
        self.assertNotIn(var, self.model.variables.values())
        self.assertEqual(var.problem, None)
        self.assertEqual(self.model.objective.direction, "max")
        self.assertEqual(
            (self.model.objective.expression - (2.0 * self.model.variables.x)).expand() - 0, 0
        )
        self.assertEqual(len(self.model.variables), 1)
        self.assertEqual(len(self.model.constraints), 1)
        self.assertEqual((self.model.variables[0].lb, self.model.variables[0].ub), (0, 10))
        self.assertEqual(self.model.constraints[0].lb, 3)
        self.assertEqual(
            (self.model.constraints[0].expression - (1.0 * self.model.variables.x)).expand() - 0, 0
        )
        # self.assertEqual(self.model.__str__(), 'Maximize\n2.0*x\nsubject to\nconstr1: 3 <= 1.0*x\nBounds\n0 <= x <= 10')

    def test_number_objective(self):
        self.model.objective = Objective(0.)
        self.assertIn('Maximize\n0', self.model.objective.__str__())

    def test_add_differing_interface_type_raises(self):
        from optlang import glpk_interface as glpk
        x, y = glpk.Variable('x'), glpk.Variable('y')
        constraint = glpk.Constraint(x + y)
        objective = glpk.Objective(1. * x)
        self.assertRaises(TypeError, self.model.add, x)
        self.assertRaises(TypeError, self.model.add, constraint)
        self.assertRaises(TypeError, self.model.add, objective)

    def test_setting_binary_bounds(self):
        self.model.variables['x'].type = 'binary'
        self.assertEqual(self.model.variables['x'].lb, 0)
        self.assertEqual(self.model.variables['x'].ub, 1)

    def test_non_integer_bound_on_integer_variable_raises(self):
        self.model.variables['x'].type = 'integer'
        self.assertRaises(ValueError, setattr, self.model.variables['x'], 'lb', 3.3)

    def test_non_0_or_1_bound_on_binary_variable_raises(self):
        self.model.variables['x'].type = 'integer'
        self.assertRaises(ValueError, setattr, self.model.variables['x'], 'lb', 3.3)

    def test_false_objective_direction_raises(self):
        self.assertRaises(ValueError, setattr, self.model.objective, 'direction', 'neither_min_nor_max')

    def test_all_entities_point_to_correct_model(self):
        for variable in self.model.variables.values():
            self.assertEqual(variable.problem, self.model)
        for constraint in self.model.constraints:
            self.assertEqual(constraint.problem, self.model)
        self.assertEqual(self.model.objective.problem, self.model)

    def test_variable_independence(self):
        model = Model()
        x = Variable('x', lb=0, ub=20)
        self.assertNotEqual(id(x), id(self.model.variables['x']))
        y = Variable('y', lb=0, ub=10)
        constr = Constraint(1. * x + y, lb=3, name="constr1")
        model.add(constr)
        self.assertNotEqual(id(self.model.variables['x']), id(model.variables['x']))
        self.assertNotEqual(id(self.model.variables['y']), id(model.variables['y']))
        self.assertNotEqual(self.model.variables['y'].problem, model)
        self.assertNotEqual(self.model.variables['x'].problem, model)
        x.lb = -10
        self.assertNotEqual(self.model.variables['x'].lb, model.variables['x'].lb)

    def test_primal_and_dual_values(self):
        model = self.model
        self.assertTrue(all([primal is None for primal in model.primal_values.values()]))
        self.assertTrue(all([constraint_primal is None for constraint_primal in model.constraint_values.values()]))
        self.assertTrue(all([sp is None for sp in model.shadow_prices.values()]))
        self.assertTrue(all([rc is None for rc in model.reduced_costs.values()]))

    def test_interface(self):
        self.assertEqual(self.model.interface, sys.modules["optlang.interface"])


class TestVariable(TestCase):
    def setUp(self):
        self.x = Variable("x")

    def test_init_variable(self):
        self.assertRaises(ValueError, Variable, '')

    def test_set_wrong_bounds_on_binary_raises(self):
        self.assertRaises(ValueError, Variable, 'x', lb=-33, ub=0.3, type='binary')
        x = Variable('x', type='binary')
        self.assertRaises(ValueError, setattr, x, 'lb', -3)
        self.assertRaises(ValueError, setattr, x, 'ub', 3)

    def test_set_wrong_bounds_on_integer_raises(self):
        self.assertRaises(ValueError, Variable, 'x', lb=-33, ub=0.3, type='integer')
        x = Variable('x', type='integer')
        self.assertRaises(ValueError, setattr, x, 'lb', -3.3)
        self.assertRaises(ValueError, setattr, x, 'ub', 3.3)

    def test_primal_and_dual(self):
        x = self.x
        self.assertTrue(x.primal is None)
        self.assertTrue(x.dual is None)

    def test_change_name(self):
        x = self.x
        x.name = "xx"
        self.assertEqual(x.name, "xx")


class TestConstraint(TestCase):
    def setUp(self):
        self.a = Variable("a")
        self.b = Variable("b")
        self.c = Variable("c")
        self.d = Variable("d")

    def test_is_linear_and_quadratic(self):
        c1 = Constraint(self.a)
        self.assertTrue(c1.is_Linear)
        self.assertFalse(c1.is_Quadratic)
        c2 = Constraint(self.a * self.b ** self.c)
        self.assertFalse(c2.is_Linear)
        self.assertFalse(c2.is_Quadratic)
        c3 = Constraint(self.a * self.b * self.c + self.d)
        self.assertFalse(c3.is_Linear)
        self.assertFalse(c3.is_Quadratic)
        c4 = Constraint(self.a + self.b ** 3)
        self.assertFalse(c4.is_Linear)
        self.assertFalse(c4.is_Quadratic)



import unittest

try:
    from optlang.inspyred_interface import Model, Objective, Variable
    import inspyred
    from inspyred import benchmarks


    def make_individual(evaluator):
        def _(candidate):
            return evaluator([candidate.values()], {})[0]

        return _


    def my_observer(population, num_generations, num_evaluations, args):
        best = max(population)
        print(('{0:6} -- {1} : {2}'.format(num_generations,
                                           best.fitness,
                                           str(best.candidate))))


    @unittest.skip("Skip these tests until module is complete")
    class RosenbrockTestCase(unittest.TestCase):
        def setUp(self):
            self.model = Model(algorithm='PSO')
            self.model.configuration.terminator = inspyred.ec.terminators.generation_termination
            x = Variable('x', lb=0, ub=2)
            y = Variable('y', lb=0, ub=2)

            rosenbrock_obj = Objective((1 - x) ** 2 + 100 * (y - y ** 2) ** 2, name="Rosenbrock function",
                                       direction='min')
            self.model.objective = rosenbrock_obj

        def test_evolutionary_strategy(self):
            self.model.configuration.max_generations = 100
            self.model.configuration.terminator = inspyred.ec.terminators.generation_termination
            self.model.configuration.observer = my_observer
            self.model.configuration.selector = inspyred.ec.selectors.tournament_selection
            final_pop = self.model.optimize()

            best = max(final_pop)
            self.assertAlmostEqual(best.fitness, 0)

        def test_pso(self):
            self.model.algorithm = 'PSO'
            self.model.configuration.max_generations = 100
            final_pop = self.model.optimize()

            best = max(final_pop)
            self.assertAlmostEqual(best.fitness, 0)

except ImportError as e:
    class TestMissingDependency(unittest.TestCase):

        @unittest.skip('Missing dependency - ' + str(e))
        def test_fail(self):
            pass



# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import islice

from optlang.exceptions import ContainerAlreadyContains


class Container(object):
    '''
    A container for objects that have a name attribute.
    Items in the Container can be accessed by index like a list or by key (the name attribute of the item).
    For convenience, an item can also be accessed as an attribute with the same name as the name attribute of the item.

    Items can be added to the Container using the list-like methods 'append' and 'extend' or by
    dict-like assignment.

    Examples
    --------
    >>> container = Container()
    >>> var1 = Variable("coffee", lb=-10, ub=10)

    The variable can be added in either of the following ways:
    >>> container.append(var1)
    >>> container.extend([var1])
    >>> container["coffee"] = var1 # Note the key must match the name attribute of the item

    The added item can be accessed in either of the following ways:
    >>> container[0]
    >>> container["coffee"]
    >>> container.coffee
    '-10 <= coffee <= 10'
    '''

    def __init__(self, iterable=()):
        self._dict = {}
        self._object_list = []
        self._indices = {}
        for item in iterable:
            try:
                item.name
            except AttributeError:
                raise AttributeError("Item " + item.__repr__() + " does not have a 'name' attribute")
            self.append(item)

    @staticmethod
    def _check_for_name_attribute(value):
        if not hasattr(value, 'name'):
            raise AttributeError('Object %s does not have a "name" attribute and cannot not be stored.' % value)

    def __len__(self):
        return len(self._dict)

    def __contains__(self, item):
        if item in self._dict:
            return True
        elif hasattr(item, "name") and item.name in self._dict:
            return item is self._dict[item.name]
        return False

    def __iter__(self):
        original_length = len(self._object_list)
        for item in self._object_list.__iter__():
            if original_length != len(self._object_list):
                raise RuntimeError("container changed size during iteration")
            yield item

    def __getitem__(self, item):
        try:
            return self._object_list[item]  # Try treating item as int or slice
        except TypeError:  # Not int or slice
            return self._dict[item]  # Treat item as key (name)

    def __setitem__(self, key, value):
        self._check_for_name_attribute(value)

        if isinstance(key, int):
            old_value = self._object_list[key]
            if old_value.name == value.name:
                self._object_list[key] = value
                self._dict[value.name] = value
            else:
                if value.name in self:
                    raise ValueError("The container already contains an object with the name " + repr(value.name))
                else:
                    self._object_list[key] = value
                    del self._dict[old_value.name]
                    self._dict[value.name] = value
        else:
            if value.name != key:
                raise ValueError("Name of item does not match key")
            try:
                old_value = self._dict[key]
            except KeyError:
                self.append(value)
            else:
                self._dict[old_value.name] = value
                self._object_list[self._indices[old_value.name]] = value

    def __delitem__(self, key):
        name = self[key].name
        index = self._indices[name]
        del self._dict[name]
        del self._object_list[index]
        del self._indices[name]
        self._reindex(index)

    def _reindex(self, start=0):
        for i, item in enumerate(islice(self._object_list, start, len(self))):
            self._indices[item.name] = start + i

    def update_key(self, key):
        item = self._dict[key]
        name = item.name
        if key != name:
            self._dict[name] = item
            del self._dict[key]
            self._indices[name] = self._indices[key]
            del self._indices[key]

    def keys(self):
        return list(item.name for item in self._object_list)

    def iterkeys(self):
        return (item.name for item in self._object_list)

    def values(self):
        return list(iter(self._object_list))

    def itervalues(self):
        return iter(self._object_list)

    def items(self):
        return ((item.name, item) for item in self._object_list)

    def iteritems(self):
        return self.items()

    def fromkeys(self, keys):
        return self.__class__((self[key] for key in keys))

    def get(self, key, default=None):
        try:
            return self[key]
        except (KeyError, IndexError):
            return default

    def clear(self):
        self._dict = {}
        self._indices = {}
        self._object_list = []

    def has_key(self, key):
        return key in self._dict

    def append(self, value):
        self._check_for_name_attribute(value)
        name = value.name
        if name in self._dict:
            raise ContainerAlreadyContains("Container '%s' already contains an object with name '%s'." % (self, value.name))
        self._indices[name] = len(self)
        self._object_list.append(value)
        self._dict[name] = value

    def extend(self, values):
        for value in values:
            self._check_for_name_attribute(value)
            if value.name in self._dict:
                raise ContainerAlreadyContains("Container '%s' already contains an object with name '%s'." % (self, value.name))
        length = len(self)
        self._object_list.extend(values)
        self._dict.update({value.name: value for value in values})
        self._indices.update({value.name: length + i for i, value in enumerate(values)})

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError("'%s' object has no attribute %s" % (self, name))

    def __getstate__(self):
        return self._object_list

    def __setstate__(self, obj_list):
        self.__init__(obj_list)

    def __dir__(self):
        attributes = list(self.__class__.__dict__.keys())
        attributes.extend(item.name for item in self._object_list)
        return attributes



# Copyright (c) 2013 Novo Nordisk Foundation Center for Biosustainability, DTU.
# See LICENSE for details.

import glob
import os
import pickle
import tarfile
import tempfile
from functools import partial

import nose
import six
from optlang.glpk_exact_interface import Model
from swiglpk import glp_create_prob, GLP_MPS_DECK, glp_read_mps, glp_get_num_cols, glp_smcp, glp_exact, \
    glp_get_status, \
    GLP_OPT, glp_get_obj_val


def test_netlib(netlib_tar_path=os.path.join(os.path.dirname(__file__), 'data/netlib_lp_problems.tar.gz')):
    """
    Test netlib with glpk interface
    """
    if six.PY3:
        nose.SkipTest('Skipping because py3')
    elif os.getenv('CI', 'false') == 'true':
        nose.SkipTest('Skipping because solving problems exactly takes very long.')
    else:
        with open(os.path.join(os.path.dirname(__file__), 'data/the_final_netlib_results.pcl'), 'rb') as fhandle:
            THE_FINAL_NETLIB_RESULTS = pickle.load(fhandle)

        # noinspection PyShadowingNames
        def read_netlib_sif_glpk(fhandle):
            tmp_file = tempfile.mktemp(suffix='.mps')
            with open(tmp_file, 'w') as tmp_handle:
                content = ''.join([str(s) for s in fhandle if str(s.strip())])
                tmp_handle.write(content)
                fhandle.close()
            problem = glp_create_prob()
            glp_read_mps(problem, GLP_MPS_DECK, None, tmp_file)
            # glp_read_mps(problem, GLP_MPS_FILE, None, tmp_file)
            return problem

        def check_dimensions(glpk_problem, model):
            """
            Tests that the glpk problem and the interface model have the same
            number of rows (constraints) and columns (variables).
            """
            assert glp_get_num_cols(glpk_problem) == len(model.variables)

        def check_objval_against_the_final_netlib_results(netlib_id, model_objval):
            print(float(THE_FINAL_NETLIB_RESULTS[netlib_id]['Objvalue']), model_objval)
            relative_error = abs(1 - (model_objval / float(THE_FINAL_NETLIB_RESULTS[netlib_id]['Objvalue'])))
            print(relative_error)
            nose.tools.assert_true(relative_error < 0.01)
            # nose.tools.assert_almost_equal(model_objval, float(THE_FINAL_NETLIB_RESULTS[netlib_id]['Objvalue']), places=4)

        tar = tarfile.open(netlib_tar_path)
        model_paths_in_tar = glob.fnmatch.filter(tar.getnames(), '*.SIF')

        for model_path_in_tar in model_paths_in_tar:
            netlib_id = os.path.basename(model_path_in_tar).replace('.SIF', '')
            # TODO: get the following problems to work
            # E226 seems to be a MPS related problem, see http://lists.gnu.org/archive/html/bug-glpk/2003-01/msg00003.html
            if netlib_id in ('AGG', 'E226', 'SCSD6', 'DFL001'):
                # def test_skip(netlib_id):
                # raise SkipTest('Skipping netlib problem %s ...' % netlib_id)
                # test_skip(netlib_id)
                # class TestWeirdNetlibProblems(unittest.TestCase):

                # @unittest.skip('Skipping netlib problem')
                # def test_fail():
                #         pass
                continue
            # TODO: For now, test only models that are covered by the final netlib results
            else:
                if netlib_id not in THE_FINAL_NETLIB_RESULTS.keys():
                    continue
                fhandle = tar.extractfile(model_path_in_tar)
                glpk_problem = read_netlib_sif_glpk(fhandle)
                model = Model(problem=glpk_problem)
                model.configuration.presolve = True
                model.configuration.verbosity = 3
                func = partial(check_dimensions, glpk_problem, model)
                func.description = "test_netlib_check_dimensions_%s (%s)" % (netlib_id, os.path.basename(str(__file__)))
                yield func

                model.optimize()
                if model.status == 'optimal':
                    model_objval = model.objective.value
                else:
                    raise Exception('No optimal solution found for netlib model %s' % netlib_id)

                func = partial(check_objval_against_the_final_netlib_results, netlib_id, model_objval)
                func.description = "test_netlib_check_objective_value__against_the_final_netlib_results_%s (%s)" % (
                    netlib_id, os.path.basename(str(__file__)))
                yield func

                if os.getenv('CI', 'false') != 'true':
                    # check that a cloned model also gives the correct result
                    model = Model.clone(model, use_json=False, use_lp=False)
                    model.optimize()
                    if model.status == 'optimal':
                        model_objval = model.objective.value
                    else:
                        raise Exception('No optimal solution found for netlib model %s' % netlib_id)

                    func = partial(check_objval_against_the_final_netlib_results, netlib_id, model_objval)
                    func.description = "test_netlib_check_objective_value__against_the_final_netlib_results_after_cloning_%s (%s)" % (
                        netlib_id, os.path.basename(str(__file__)))
                    yield func


if __name__ == '__main__':
    # tar = tarfile.open('data/netlib_lp_problems.tar.gz')
    # model_paths_in_tar = glob.fnmatch.filter(tar.getnames(), '*.SIF')
    # fhandle = tar.extractfile('./netlib/ADLITTLE.SIF')
    # glpk_problem = read_netlib_sif_glpk(fhandle)
    # glp_simplex(glpk_problem, None)
    # print glp_get_obj_val(glpk_problem)
    # print glpk_problem
    # fhandle = tar.extractfile('./netlib/ADLITTLE.SIF')
    # glpk_problem = read_netlib_sif_glpk(fhandle)
    # model = Model(problem=glpk_problem)
    # glp_simplex(glpk_problem, None)
    # model.optimize()
    # print model.objective.value
    # print model
    # test_netlib().next()
    nose.runmodule()



# Copyright (c) 2013 Novo Nordisk Foundation Center for Biosustainability, DTU.
# See LICENSE for details.

import pickle
import unittest

from optlang.interface import Variable, Constraint, Model


class VariableTestCase(unittest.TestCase):
    def setUp(self):
        self.var = Variable('x')

    def test_white_space_name_raises(self):
        with self.assertRaises(ValueError):
            Variable('White Space')

    def test_init(self):
        self.assertEqual(self.var.name, 'x')
        self.assertEqual(self.var.lb, None)
        self.assertEqual(self.var.ub, None)
        self.assertEqual(self.var.type, 'continuous')
        self.assertEqual(self.var.problem, None)

        y = Variable('y', lb=-10)
        self.assertEqual(y.lb, -10)
        self.assertEqual(y.ub, None)

        u = Variable('u', ub=-10)
        self.assertEqual(u.lb, None)
        self.assertEqual(u.ub, -10)

    def test_change_lb(self):
        self.var.lb = 666
        self.assertEqual(self.var.lb, 666)
        self.var.lb = None
        self.assertEqual(self.var.lb, None)

    def test_change_ub(self):
        self.var.ub = -465.3323
        self.assertEqual(self.var.ub, -465.3323)
        self.var.ub = None
        self.assertEqual(self.var.ub, None)

    def test_change_type(self):
        self.var.type = 'integer'
        self.assertEqual(self.var.type, 'integer')
        self.var.type = 'binary'
        self.assertEqual(self.var.type, 'binary')
        self.var.type = 'continuous'
        self.assertEqual(self.var.type, 'continuous')
        self.assertRaises(ValueError, setattr, self.var, 'type', 'unicorn')

    def test_change_problem(self):
        model = Model()
        self.var.problem = model
        self.assertEqual(self.var.problem, model)

    def test_lb_greater_than_ub_raises(self):
        self.var.lb = -10
        self.var.ub = 10
        self.assertRaises(Exception, setattr, self.var, 'lb', 20)

    def test_ub_smaller_than_lb_raises(self):
        self.var.lb = 0
        self.var.ub = 10
        self.assertRaises(Exception, setattr, self.var, 'ub', -20)

    def test_pickle_ability(self):
        var = Variable('name', type='binary')
        pickle_var = pickle.loads(pickle.dumps(var))
        keys = var.__dict__.keys()
        self.assertEqual([getattr(var, k) for k in keys], [getattr(pickle_var, k) for k in keys])


# noinspection PyPep8Naming
class ConstraintTestCase(unittest.TestCase):
    def setUp(self):
        self.x = Variable('x', type='binary')
        self.y = Variable('y', lb=-181133.3, ub=12000., type='continuous')
        self.z = Variable('z', lb=0., ub=3., type='integer')

    def test_is_Linear(self):
        constraint = Constraint(
            0.3 * self.x + 0.4 * self.y + self.y + 66. * self.z, lb=-100, ub=0., name='linear_constraint')
        self.assertTrue(constraint.is_Linear)
        self.assertFalse(constraint.is_Quadratic)

    def test_is_Quadratic_pow(self):
        constraint = Constraint(
            0.3 * self.x + 0.4 * self.y ** 2 + self.y + 66. * self.z, lb=-100, ub=0., name='quad_pow_constraint')
        self.assertFalse(constraint.is_Linear)
        self.assertTrue(constraint.is_Quadratic)

    def test_is_Quadratic_xy(self):
        constraint = Constraint(
            0.3 * self.x + 0.4 * self.y * self.x + self.y + 66. * self.z, lb=-100, ub=0., name='quad_xy_constraint')
        self.assertFalse(constraint.is_Linear)
        self.assertTrue(constraint.is_Quadratic)
        constraint = Constraint(
            0.3 * self.x ** 2 + 0.4 * self.y + self.y + 66. * self.z, lb=-100, ub=0., name='quad_xy_constraint')
        self.assertFalse(constraint.is_Linear)
        self.assertTrue(constraint.is_Quadratic)

    def test_catching_nonlinear_expressions(self):
        constraint = Constraint(
            0.3 * self.x ** 3 + 0.4 * self.y * self.x + self.y + 66. * self.z, lb=-100, ub=0.,
            name='nonlinear_constraint')
        self.assertFalse(constraint.is_Linear)
        self.assertFalse(constraint.is_Quadratic)
        constraint = Constraint(
            0.3 * self.x ** self.y + 0.4 * self.y * self.x + self.y + 66. * self.z, lb=-100, ub=0.,
            name='nonlinear_constraint')
        self.assertFalse(constraint.is_Linear)
        self.assertFalse(constraint.is_Quadratic)

    def test_lonely_coefficient_and_no_bounds_raises(self):
        self.assertRaises(ValueError, Constraint, self.x + 0.3, name='lonely_coeff_constraint')

    def test_canonicalization_with_lb(self):
        constraint = Constraint(-20 + self.x + 3, lb=-666)
        self.assertEqual(constraint.lb, -649)
        self.assertEqual(constraint.ub, None)
        self.assertEqual(constraint.expression, self.x)

    def test_canonicalization_with_ub(self):
        constraint = Constraint(-20 + self.x + 3, ub=-666)
        self.assertEqual(constraint.ub, -649)
        self.assertEqual(constraint.lb, None)
        self.assertEqual(constraint.expression, self.x)

    def test_canonicalization_with_lb_and_ub(self):
        constraint = Constraint(-20 + self.x + 3, ub=666, lb=-666)
        self.assertEqual(constraint.lb, -649)
        self.assertEqual(constraint.ub, 683)
        self.assertEqual(constraint.expression, self.x)

        # def test_pickle_ability(self):
        # constraint = Constraint(
        #         0.3*self.x**self.y + 0.4*self.y*self.x + self.y + 66.*self.z, lb=-100, ub=0., name='nonlinear_constraint')
        #     pickle_constraint = pickle.loads(pickle.dumps(constraint))
        #     keys = constraint.__dict__.keys()
        #     self.assertEqual([getattr(constraint, k) for k in keys], [getattr(pickle_constraint, k) for k in keys])


if __name__ == '__main__':
    import nose

    nose.runmodule()



# Copyright (c) 2013 Novo Nordisk Foundation Center for Biosustainability, DTU.
# See LICENSE for details.

import os
import random
import unittest
import json

import nose

import copy
import pickle

from optlang import glpk_exact_interface
from optlang.tests import abstract_test_cases

from optlang import glpk_interface
from optlang.util import glpk_read_cplex
from swiglpk import glp_get_num_rows, glp_get_col_name, glp_get_num_cols, glp_get_prob_name, glp_get_row_name, \
    glp_get_col_kind, glp_find_col, intArray, doubleArray, glp_get_mat_row, glp_get_row_type, glp_get_row_lb, \
    glp_get_row_ub, glp_get_obj_coef, GLP_UP, GLP_DB, GLP_LO, GLP_CV, GLP_IV, GLP_FX, GLP_FR, glp_get_col_lb, \
    glp_get_col_ub, glp_get_obj_dir, GLP_MIN, GLP_MAX


random.seed(666)
TESTMODELPATH = os.path.join(os.path.dirname(__file__), 'data/model.lp')
TESTMILPMODELPATH = os.path.join(os.path.dirname(__file__), 'data/simple_milp.lp')
ECOLI_TEST = os.path.join(os.path.dirname(__file__), 'data/coli_core.json')


class VariableTestCase(abstract_test_cases.AbstractVariableTestCase):
    interface = glpk_exact_interface

    def test_get_primal(self):
        self.assertEqual(self.var.primal, None)
        model = glpk_interface.Model(problem=glpk_read_cplex(TESTMODELPATH))
        model.optimize()
        for i, j in zip([var.primal for var in model.variables],
                        [0.8739215069684306, -16.023526143167608, 16.023526143167604, -14.71613956874283,
                         14.71613956874283, 4.959984944574658, 4.959984944574657, 4.959984944574658,
                         3.1162689467973905e-29, 2.926716099010601e-29, 0.0, 0.0, -6.112235045340358e-30,
                         -5.6659435396316186e-30, 0.0, -4.922925402711085e-29, 0.0, 9.282532599166613, 0.0,
                         6.00724957535033, 6.007249575350331, 6.00724957535033, -5.064375661482091, 1.7581774441067828,
                         0.0, 7.477381962160285, 0.0, 0.22346172933182767, 45.514009774517454, 8.39, 0.0,
                         6.007249575350331, 0.0, -4.541857463865631, 0.0, 5.064375661482091, 0.0, 0.0,
                         2.504309470368734, 0.0, 0.0, -22.809833310204958, 22.809833310204958, 7.477381962160285,
                         7.477381962160285, 1.1814980932459636, 1.496983757261567, -0.0, 0.0, 4.860861146496815, 0.0,
                         0.0, 0.0, 0.0, 0.0, 0.0, 5.064375661482091, 0.0, 5.064375661482091, 0.0, 0.0,
                         1.496983757261567, 10.000000000000002, -10.0, 0.0, 0.0, 0.0, 0.0, 0.0, -29.175827135565804,
                         43.598985311997524, 29.175827135565804, 0.0, 0.0, 0.0, -1.2332237321082153e-29,
                         3.2148950476847613, 38.53460965051542, 5.064375661482091, 0.0, -1.2812714099825612e-29,
                         -1.1331887079263237e-29, 17.530865429786694, 0.0, 0.0, 0.0, 4.765319193197458,
                         -4.765319193197457, 21.79949265599876, -21.79949265599876, -3.2148950476847613, 0.0,
                         -2.281503094067127, 2.6784818505075303, 0.0]):
            self.assertAlmostEqual(i, j)

    def test_changing_variable_names_is_reflected_in_the_solver(self):
        model = self.interface.Model(problem=glpk_read_cplex(TESTMODELPATH))
        for i, variable in enumerate(model.variables):
            variable.name = "var" + str(i)
            self.assertEqual(variable.name, "var" + str(i))
            self.assertEqual(glp_get_col_name(model.problem, variable._index), "var" + str(i))

    def test_change_type(self):
        self.var.type = "continuous"
        self.assertRaises(ValueError, setattr, self.var, "type", "integer")
        self.assertRaises(ValueError, setattr, self.var, "type", "binary")

    def test_set_wrong_type_raises(self):
        self.assertRaises(ValueError, self.interface.Variable, name="test", type="mayo")
        self.assertRaises(Exception, setattr, self.var, 'type', 'ketchup')
        self.model.add(self.var)
        self.model.update()
        self.assertRaises(ValueError, setattr, self.var, "type", "mustard")


class ConstraintTestCase(abstract_test_cases.AbstractConstraintTestCase):
    interface = glpk_exact_interface

    def test_get_primal(self):
        pass

    @unittest.skip("NA")
    def test_indicator_constraint_support(self):
        pass


class ObjectiveTestCase(abstract_test_cases.AbstractObjectiveTestCase):
    interface = glpk_exact_interface

    def setUp(self):
        with open(ECOLI_TEST) as infile:
            self.model = self.interface.Model.from_json(json.load(infile))
        self.obj = self.model.objective

    def test_change_direction(self):
        self.obj.direction = "min"
        self.assertEqual(self.obj.direction, "min")
        self.assertEqual(glp_get_obj_dir(self.model.problem), GLP_MIN)

        self.obj.direction = "max"
        self.assertEqual(self.obj.direction, "max")
        self.assertEqual(glp_get_obj_dir(self.model.problem), GLP_MAX)


class ConfigurationTestCase(abstract_test_cases.AbstractConfigurationTestCase):
    interface = glpk_exact_interface


class ModelTestCase(abstract_test_cases.AbstractModelTestCase):
    interface = glpk_exact_interface

    def test_pickle_ability(self):
        self.model.optimize()
        value = self.model.objective.value
        pickle_string = pickle.dumps(self.model)
        from_pickle = pickle.loads(pickle_string)
        from_pickle.optimize()
        self.assertAlmostEqual(value, from_pickle.objective.value)
        self.assertEqual([(var.lb, var.ub, var.name, var.type) for var in from_pickle.variables.values()],
                         [(var.lb, var.ub, var.name, var.type) for var in self.model.variables.values()])
        self.assertEqual([(constr.lb, constr.ub, constr.name) for constr in from_pickle.constraints],
                         [(constr.lb, constr.ub, constr.name) for constr in self.model.constraints])

    def test_config_gets_copied_too(self):
        self.assertEqual(self.model.configuration.verbosity, 0)
        self.model.configuration.verbosity = 3
        model_copy = copy.copy(self.model)
        self.assertEqual(model_copy.configuration.verbosity, 3)

    def test_init_from_existing_problem(self):
        inner_prob = self.model.problem
        self.assertEqual(len(self.model.variables), glp_get_num_cols(inner_prob))
        self.assertEqual(len(self.model.constraints), glp_get_num_rows(inner_prob))
        self.assertEqual(self.model.variables.keys(),
                         [glp_get_col_name(inner_prob, i) for i in range(1, glp_get_num_cols(inner_prob) + 1)])
        self.assertEqual(self.model.constraints.keys(),
                         [glp_get_row_name(inner_prob, j) for j in range(1, glp_get_num_rows(inner_prob) + 1)])

    def test_add_non_cplex_conform_variable(self):
        var = self.interface.Variable('12x!!@#5_3', lb=-666, ub=666)
        self.assertEqual(var._index, None)
        self.model.add(var)
        self.assertTrue(var in self.model.variables.values())
        self.assertEqual(var.name, glp_get_col_name(self.model.problem, var._index))
        self.assertEqual(self.model.variables['12x!!@#5_3'].lb, -666)
        self.assertEqual(self.model.variables['12x!!@#5_3'].ub, 666)
        repickled = pickle.loads(pickle.dumps(self.model))
        var_from_pickle = repickled.variables['12x!!@#5_3']
        self.assertEqual(var_from_pickle.name, glp_get_col_name(repickled.problem, var_from_pickle._index))

    def test_change_of_constraint_is_reflected_in_low_level_solver(self):
        x = self.interface.Variable('x', lb=-83.3, ub=1324422.)
        y = self.interface.Variable('y', lb=-181133.3, ub=12000.)
        constraint = self.interface.Constraint(0.3 * x + 0.4 * y, lb=-100, name='test')
        self.assertEqual(constraint._index, None)
        self.model.add(constraint)
        self.assertEqual(
            (self.model.constraints["test"].expression - (0.4 * y + 0.3 * x)).expand(), 0
        )
        self.assertEqual(self.model.constraints["test"].lb, -100)

        self.assertEqual(constraint._index, 73)
        z = self.interface.Variable('z', lb=3, ub=10, type='continuous')
        self.assertEqual(z._index, None)
        constraint += 77. * z
        self.assertEqual(z._index, 98)

        self.assertEqual(
            (self.model.constraints["test"].expression - (0.4 * y + 0.3 * x + 77.0 * z)).expand(), 0
        )
        self.assertEqual(self.model.constraints["test"].lb, -100)

        self.assertEqual(constraint._index, 73)

    def test_constraint_set_problem_to_None_caches_the_latest_expression_from_solver_instance(self):
        x = self.interface.Variable('x', lb=-83.3, ub=1324422.)
        y = self.interface.Variable('y', lb=-181133.3, ub=12000.)
        constraint = self.interface.Constraint(0.3 * x + 0.4 * y, lb=-100, name='test')
        self.model.add(constraint)
        z = self.interface.Variable('z', lb=2, ub=5, type='continuous')
        constraint += 77. * z
        self.model.remove(constraint)
        self.assertEqual(
            (constraint.expression - (0.4 * y + 0.3 * x + 77.0 * z)).expand() - 0, 0
        )
        self.assertEqual(constraint.lb, -100)
        self.assertEqual(constraint.ub, None)

    def test_change_of_objective_is_reflected_in_low_level_solver(self):
        x = self.interface.Variable('x', lb=-83.3, ub=1324422.)
        y = self.interface.Variable('y', lb=-181133.3, ub=12000.)
        objective = self.interface.Objective(0.3 * x + 0.4 * y, name='test', direction='max')
        self.model.objective = objective

        self.assertEqual(
            (self.model.objective.expression - (0.4 * y + 0.3 * x)).expand() - 0, 0
        )
        self.assertEqual(self.model.objective.direction, "max")

        self.assertEqual(glp_get_obj_coef(self.model.problem, x._index), 0.3)
        self.assertEqual(glp_get_obj_coef(self.model.problem, y._index), 0.4)
        for i in range(1, glp_get_num_cols(self.model.problem) + 1):
            if i != x._index and i != y._index:
                self.assertEqual(glp_get_obj_coef(self.model.problem, i), 0)
        z = self.interface.Variable('z', lb=4, ub=4, type='continuous')
        self.model.objective += 77. * z

        self.assertEqual(
            (self.model.objective.expression - (0.4 * y + 0.3 * x + 77.0 * z)).expand() - 0, 0
        )
        self.assertEqual(self.model.objective.direction, "max")

        self.assertEqual(glp_get_obj_coef(self.model.problem, x._index), 0.3)
        self.assertEqual(glp_get_obj_coef(self.model.problem, y._index), 0.4)
        self.assertEqual(glp_get_obj_coef(self.model.problem, z._index), 77.)
        for i in range(1, glp_get_num_cols(self.model.problem) + 1):
            if i != x._index and i != y._index and i != z._index:
                self.assertEqual(glp_get_obj_coef(self.model.problem, i), 0)

    def test_change_variable_bounds(self):
        inner_prob = self.model.problem
        inner_problem_bounds = [(glp_get_col_lb(inner_prob, i), glp_get_col_ub(inner_prob, i)) for i in
                                range(1, glp_get_num_cols(inner_prob) + 1)]
        bounds = [(var.lb, var.ub) for var in self.model.variables.values()]
        self.assertEqual(bounds, inner_problem_bounds)
        for var in self.model.variables.values():
            var.lb = random.uniform(-1000, 1000)
            var.ub = random.uniform(var.lb, 1000)
        inner_problem_bounds_new = [(glp_get_col_lb(inner_prob, i), glp_get_col_ub(inner_prob, i)) for i in
                                    range(1, glp_get_num_cols(inner_prob) + 1)]
        bounds_new = [(var.lb, var.ub) for var in self.model.variables.values()]
        self.assertNotEqual(bounds, bounds_new)
        self.assertNotEqual(inner_problem_bounds, inner_problem_bounds_new)
        self.assertEqual(bounds_new, inner_problem_bounds_new)

    def test_change_constraint_bounds(self):
        inner_prob = self.model.problem
        inner_problem_bounds = [(glp_get_row_lb(inner_prob, i), glp_get_row_ub(inner_prob, i)) for i in
                                range(1, glp_get_num_rows(inner_prob) + 1)]
        bounds = [(constr.lb, constr.ub) for constr in self.model.constraints]
        self.assertEqual(bounds, inner_problem_bounds)
        for constr in self.model.constraints:
            constr.lb = random.uniform(-1000, constr.ub)
            constr.ub = random.uniform(constr.lb, 1000)
        inner_problem_bounds_new = [(glp_get_row_lb(inner_prob, i), glp_get_row_ub(inner_prob, i)) for i in
                                    range(1, glp_get_num_rows(inner_prob) + 1)]
        bounds_new = [(constr.lb, constr.ub) for constr in self.model.constraints]
        self.assertNotEqual(bounds, bounds_new)
        self.assertNotEqual(inner_problem_bounds, inner_problem_bounds_new)
        self.assertEqual(bounds_new, inner_problem_bounds_new)

    def test_iadd_objective(self):
        v2, v3 = self.model.variables.values()[1:3]
        self.model.objective += 2. * v2 - 3. * v3
        obj_coeff = list()
        for i in range(len(self.model.variables)):
            obj_coeff.append(glp_get_obj_coef(self.model.problem, i))
        self.assertEqual(obj_coeff,
                         [0.0, 1.0, 2.0, -3.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0]
                         )

    def test_imul_objective(self):
        self.model.objective *= 2.
        obj_coeff = list()
        for i in range(len(self.model.variables)):
            obj_coeff.append(glp_get_obj_coef(self.model.problem, i))
        self.assertEqual(obj_coeff,
                         [0.0, 2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0]
                         )

    def test_set_copied_objective(self):
        obj_copy = copy.copy(self.model.objective)
        self.model.objective = obj_copy
        self.assertEqual(
            (self.model.objective.expression - 1.0 * self.model.variables.R_Biomass_Ecoli_core_w_GAM).expand() - 0, 0
        )
        self.assertEqual(self.model.objective.direction, "max")

    def test_timeout(self):
        self.model.configuration.timeout = 0
        status = self.model.optimize()
        self.assertEqual(status, 'time_limit')

    def test_set_linear_coefficients_objective(self):
        self.model.objective.set_linear_coefficients({self.model.variables.R_TPI: 666.})
        self.assertEqual(glp_get_obj_coef(self.model.problem, self.model.variables.R_TPI._index), 666.)

    def test_set_linear_coefficients_constraint(self):
        constraint = self.model.constraints.M_atp_c
        constraint.set_linear_coefficients({self.model.variables.R_Biomass_Ecoli_core_w_GAM: 666.})
        num_cols = glp_get_num_cols(self.model.problem)
        ia = intArray(num_cols + 1)
        da = doubleArray(num_cols + 1)
        index = constraint._index
        num = glp_get_mat_row(self.model.problem, index, ia, da)
        for i in range(1, num + 1):
            col_name = glp_get_col_name(self.model.problem, ia[i])
            if col_name == 'R_Biomass_Ecoli_core_w_GAM':
                self.assertEqual(da[i], 666.)

    def test_remove_constraints(self):
        x = self.interface.Variable('x', type='continuous')
        y = self.interface.Variable('y', lb=-181133.3, ub=12000., type='continuous')
        z = self.interface.Variable('z', lb=4, ub=4, type='continuous')
        constr1 = self.interface.Constraint(0.3 * x + 0.4 * y + 66. * z, lb=-100, ub=0., name='test')
        self.assertEqual(constr1.problem, None)
        self.model.add(constr1)
        self.model.update()
        self.assertEqual(constr1.problem, self.model)
        self.assertIn(constr1, self.model.constraints)
        self.model.remove(constr1.name)
        self.model.update()
        self.assertEqual(constr1.problem, None)
        self.assertNotIn(constr1, self.model.constraints)

    def test_add_constraints(self):
        x = self.interface.Variable('x', lb=0, ub=1, type='continuous')
        y = self.interface.Variable('y', lb=-181133.3, ub=12000., type='continuous')
        z = self.interface.Variable('z', lb=0., ub=10., type='continuous')
        constr1 = self.interface.Constraint(0.3 * x + 0.4 * y + 66. * z, lb=-100, ub=0., name='test')
        constr2 = self.interface.Constraint(2.333 * x + y + 3.333, ub=100.33, name='test2')
        constr3 = self.interface.Constraint(2.333 * x + y + z, lb=-300)
        constr4 = self.interface.Constraint(x, lb=-300, ub=-300)
        constr5 = self.interface.Constraint(3 * x)
        self.model.add(constr1)
        self.model.add(constr2)
        self.model.add(constr3)
        self.model.add([constr4, constr5])
        self.assertIn(constr1.name, self.model.constraints)
        self.assertIn(constr2.name, self.model.constraints)
        self.assertIn(constr3.name, self.model.constraints)
        self.assertIn(constr4.name, self.model.constraints)
        self.assertIn(constr5.name, self.model.constraints)

    def test_add_nonlinear_constraint_raises(self):
        x = self.interface.Variable('x', type='continuous')
        y = self.interface.Variable('y', lb=-181133.3, ub=12000., type='continuous')
        z = self.interface.Variable('z', lb=3, ub=3, type='continuous')
        with self.assertRaises(ValueError):
            constraint = self.interface.Constraint(0.3 * x + 0.4 * y ** x + 66. * z, lb=-100, ub=0., name='test')
            self.model.add(constraint)
            self.model.update()

    def test_change_variable_type(self):
        self.assertRaises(ValueError, setattr, self.model.variables[-1], "type", "integer")

    def test_add_integer_var(self):
        self.assertRaises(ValueError, self.interface.Variable, 'int_var', lb=-13, ub=499., type='integer')

    def test_is_integer(self):
        self.skipTest("No integers with glpk_exact")

    def test_binary_variables(self):
        self.skipTest("No integers with glpk_exact")

    def test_implicitly_convert_milp_to_lp(self):
        self.skipTest("No integers with glpk_exact")

    def test_optimize_milp(self):
        self.skipTest("No integers with glpk_exact")

    def test_integer_variable_dual(self):
        self.skipTest("No integers with glpk_exact")

    def test_integer_constraint_dual(self):
        self.skipTest("No integers with glpk_exact")

    def test_integer_batch_duals(self):
        self.skipTest("No integers with glpk_exact")


if __name__ == '__main__':
    nose.runmodule()



# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Solver interface for the Gurobi (MI)LP/QP solver.

To use this interface, install the gurobi solver and the bundled python interface
and make sure that 'import gurobipy' runs without error.
"""

import logging

log = logging.getLogger(__name__)

import os
import six
from functools import partial
from optlang import interface
from optlang.util import inheritdocstring, TemporaryFilename
from optlang.expression_parsing import parse_optimization_expression
from optlang import symbolics

import gurobipy

_GUROBI_STATUS_TO_STATUS = {
    gurobipy.GRB.LOADED: interface.LOADED,
    gurobipy.GRB.OPTIMAL: interface.OPTIMAL,
    gurobipy.GRB.INFEASIBLE: interface.INFEASIBLE,
    gurobipy.GRB.INF_OR_UNBD: interface.INFEASIBLE_OR_UNBOUNDED,
    gurobipy.GRB.UNBOUNDED: interface.UNBOUNDED,
    gurobipy.GRB.CUTOFF: interface.CUTOFF,
    gurobipy.GRB.ITERATION_LIMIT: interface.ITERATION_LIMIT,
    gurobipy.GRB.NODE_LIMIT: interface.NODE_LIMIT,
    gurobipy.GRB.TIME_LIMIT: interface.TIME_LIMIT,
    gurobipy.GRB.SOLUTION_LIMIT: interface.SOLUTION_LIMIT,
    gurobipy.GRB.INTERRUPTED: interface.INTERRUPTED,
    gurobipy.GRB.NUMERIC: interface.NUMERIC,
    gurobipy.GRB.SUBOPTIMAL: interface.SUBOPTIMAL,
    gurobipy.GRB.INPROGRESS: interface.INPROGRESS
}

_LP_METHODS = {"auto": -1, "primal": 0, "dual": 1, "barrier": 2, "concurrent": 3, "deterministic_concurrent": 4}
_REVERSE_LP_METHODS = {v: k for k, v in _LP_METHODS.items()}
_QP_METHODS = {"auto": -1, "primal": 0, "dual": 1, "barrier": 2}
_REVERSE_QP_METHODS = {v: k for k, v in _QP_METHODS.items()}

_VTYPE_TO_GUROBI_VTYPE = {'continuous': gurobipy.GRB.CONTINUOUS, 'integer': gurobipy.GRB.INTEGER,
                          'binary': gurobipy.GRB.BINARY}
_GUROBI_VTYPE_TO_VTYPE = dict((val, key) for key, val in _VTYPE_TO_GUROBI_VTYPE.items())


def _constraint_lb_and_ub_to_gurobi_sense_rhs_and_range_value(lb, ub):
    """Helper function used by Constraint and Model"""
    if lb is None and ub is None:
        raise Exception("Free constraint ...")
    elif lb is None:
        sense = '<'
        rhs = float(ub)
        range_value = 0.
    elif ub is None:
        sense = '>'
        rhs = float(lb)
        range_value = 0.
    elif lb == ub:
        sense = '='
        rhs = float(lb)
        range_value = 0.
    elif lb > ub:
        raise ValueError("Lower bound is larger than upper bound.")
    else:
        sense = '='
        rhs = float(lb)
        range_value = float(ub - lb)
    return sense, rhs, range_value


@six.add_metaclass(inheritdocstring)
class Variable(interface.Variable):
    def __init__(self, name, *args, **kwargs):
        super(Variable, self).__init__(name, **kwargs)
        self._original_name = name  # due to bug with getVarByName ... TODO: file bug report on gurobi google group

    @property
    def _internal_variable(self):
        if getattr(self, 'problem', None) is not None:

            internal_variable = self.problem.problem.getVarByName(self._original_name)

            assert internal_variable is not None
            # if internal_variable is None:
            #     raise Exception('Cannot find variable {}')
            return internal_variable
        else:
            return None

    def _set_variable_bounds_on_problem(self, var_lb, var_ub):
        lb = [
            (var.name, -gurobipy.GRB.INFINITY) if val is None else (var.name, val) for var, val in var_lb
            ]
        if len(lb) > 0:
            self.problem.variables.set_lower_bounds(lb)
        ub = [
            (var.name, gurobipy.GRB.INFINITY) if val is None else (var.name, val) for var, val in var_ub
            ]
        if len(ub) > 0:
            self.problem.variables.set_upper_bounds(ub)

    @interface.Variable.lb.setter
    def lb(self, value):
        super(Variable, self.__class__).lb.fset(self, value)
        if value is None:
            value = -gurobipy.GRB.INFINITY
        if self.problem:
            self._internal_variable.setAttr('LB', value)
            self.problem.problem.update()

    @interface.Variable.ub.setter
    def ub(self, value):
        super(Variable, self.__class__).ub.fset(self, value)
        if value is None:
            value = gurobipy.GRB.INFINITY
        if self.problem:
            self._internal_variable.setAttr('UB', value)
            self.problem.problem.update()

    def set_bounds(self, lb, ub):
        super(Variable, self).set_bounds(lb, ub)
        if self.problem:
            if lb is None:
                lb = -gurobipy.GRB.INFINITY
            if ub is None:
                ub = gurobipy.GRB.INFINITY
            var = self._internal_variable
            var.setAttr("LB", lb)
            var.setAttr("UB", ub)


    @interface.Variable.type.setter
    def type(self, value):
        super(Variable, Variable).type.fset(self, value)
        if self.problem:
            return self._internal_variable.setAttr('VType', _VTYPE_TO_GUROBI_VTYPE[value])

    def _get_primal(self):
        return self._internal_variable.getAttr('X')

    @property
    def dual(self):
        if self.problem:
            if self.problem.is_integer:
                raise ValueError("Dual values are not well-defined for integer problems")
            return self._internal_variable.getAttr('RC')
        else:
            return None

    @interface.Variable.name.setter
    def name(self, value):
        internal_var = self._internal_variable
        super(Variable, Variable).name.fset(self, value)
        if internal_var is not None:
            internal_var.setAttr('VarName', value)
            self.problem.problem.update()


@six.add_metaclass(inheritdocstring)
class Constraint(interface.Constraint):
    _INDICATOR_CONSTRAINT_SUPPORT = False

    def __init__(self, expression, *args, **kwargs):
        super(Constraint, self).__init__(expression, *args, **kwargs)

    def set_linear_coefficients(self, coefficients):
        if self.problem is not None:
            self.problem.update()
            grb_constraint = self.problem.problem.getConstrByName(self.name)
            for var, coeff in six.iteritems(coefficients):
                self.problem.problem.chgCoeff(grb_constraint, self.problem.problem.getVarByName(var.name), float(coeff))
            self.problem.update()
        else:
            raise Exception("Can't change coefficients if constraint is not associated with a model.")

    def get_linear_coefficients(self, variables):
        if self.problem is not None:
            self.problem.update()
            grb_constraint = self.problem.problem.getConstrByName(self.name)
            return {v: self.problem.problem.getCoeff(grb_constraint, v._internal_variable) for v in variables}
        else:
            raise Exception("Can't get coefficients from solver if constraint is not in a model")

    @property
    def _internal_constraint(self):
        if self.problem is not None:
            return self.problem.problem.getConstrByName(self.name)

    def _get_expression(self):
        if self.problem is not None:
            gurobi_problem = self.problem.problem
            gurobi_constraint = self._internal_constraint
            row = gurobi_problem.getRow(gurobi_constraint)
            terms = []
            for i in range(row.size()):
                internal_var_name = row.getVar(i).VarName
                if internal_var_name == self.name + '_aux':
                    continue
                variable = self.problem._variables[internal_var_name]
                coeff = symbolics.Real(row.getCoeff(i))
                terms.append(symbolics.mul((coeff, variable)))
            self._expression = symbolics.add(terms)
        return self._expression

    def __str__(self):
        if self.problem is not None:
            self.problem.problem.update()
        return super(Constraint, self).__str__()

    @property
    def problem(self):
        return self._problem

    @problem.setter
    def problem(self, value):
        if value is None:
            # Update expression from solver instance one last time
            self._get_expression()
            self._problem = None
        else:
            self._problem = value

    @property
    def primal(self):
        if self.problem is not None:
            aux_var = self.problem.problem.getVarByName(self._internal_constraint.getAttr('ConstrName') + '_aux')
            if aux_var is not None:
                aux_val = aux_var.X
            else:
                aux_val = 0
            return (self._internal_constraint.RHS -
                    self._internal_constraint.Slack +
                    aux_val)
            # return self._round_primal_to_bounds(primal_from_solver)  # Test assertions fail
            # return primal_from_solver
        else:
            return None

    @property
    def dual(self):
        if self.problem is not None:
            if self.problem.is_integer:
                raise ValueError("Dual values are not well-defined for integer problems")
            return self._internal_constraint.Pi
        else:
            return None

    @interface.Constraint.name.setter
    def name(self, value):
        internal_const = self._internal_constraint
        super(Constraint, Constraint).name.fset(self, value)
        if internal_const is not None:
            internal_const.setAttr('ConstrName', value)
            self.problem.problem.update()

    def _set_constraint_bounds(self, sense, rhs, range_value):
        gurobi_constraint = self.problem.problem.getConstrByName(self.name)
        gurobi_constraint.setAttr('Sense', sense)
        gurobi_constraint.setAttr('RHS', rhs)

        aux_var = self.problem.problem.getVarByName(gurobi_constraint.getAttr('ConstrName') + '_aux')
        if range_value == 0.:
            if aux_var is not None:
                aux_var.setAttr("UB", 0)
        if range_value != 0:
            if aux_var is None:
                aux_var = self.problem.problem.addVar(name=gurobi_constraint.getAttr('ConstrName') + '_aux', lb=0,
                                                      ub=range_value)
                row = self.problem.problem.getRow(gurobi_constraint)
                updated_row = row - aux_var
                self.problem.problem.remove(gurobi_constraint)
                self.problem.problem.update()
                self.problem.problem.addConstr(updated_row, sense, rhs, self.name)
            aux_var.setAttr("UB", range_value)
        self.problem.update()

    @interface.Constraint.lb.setter
    def lb(self, value):
        super(Constraint, Constraint).lb.fset(self, value)
        if getattr(self, 'problem', None) is not None:
            if value is not None and self.ub is not None and value > self.ub:
                raise ValueError(
                    "Lower bound %f is larger than upper bound %f in constraint %s" %
                    (value, self.ub, self)
                )
            sense, rhs, range_value = _constraint_lb_and_ub_to_gurobi_sense_rhs_and_range_value(value, self.ub)
            self._set_constraint_bounds(sense, rhs, range_value)

    @interface.Constraint.ub.setter
    def ub(self, value):
        super(Constraint, Constraint).ub.fset(self, value)
        if getattr(self, 'problem', None) is not None:
            if value is not None and self.lb is not None and value < self.lb:
                raise ValueError(
                    "Upper bound %f is less than lower bound %f in constraint %s" %
                    (value, self.lb, self)
                )
            sense, rhs, range_value = _constraint_lb_and_ub_to_gurobi_sense_rhs_and_range_value(self.lb, value)
            self._set_constraint_bounds(sense, rhs, range_value)

    def __iadd__(self, other):
        if self.problem is not None:
            problem_reference = self.problem
            self.problem._remove_constraint(self)
            super(Constraint, self).__iadd__(other)
            problem_reference._add_constraint(self, sloppy=False)
        else:
            super(Constraint, self).__iadd__(other)
        return self


@six.add_metaclass(inheritdocstring)
class Objective(interface.Objective):
    def __init__(self, expression, sloppy=False, *args, **kwargs):
        super(Objective, self).__init__(expression, *args, sloppy=sloppy, **kwargs)
        self._expression_expired = False
        if not (sloppy or self.is_Linear or self.is_Quadratic):
            raise ValueError("The given objective is invalid. Must be linear or quadratic.")

    @property
    def value(self):
        if getattr(self, "problem", None) is not None:
            gurobi_problem = self.problem.problem
            try:
                return gurobi_problem.getAttr("ObjVal") + getattr(self.problem, "_objective_offset", 0)
            except (gurobipy.GurobiError, AttributeError):  # TODO: let this check the actual error message
                return None
        else:
            return None

    @interface.Objective.direction.setter
    def direction(self, value):
        super(Objective, Objective).direction.__set__(self, value)
        if getattr(self, 'problem', None) is not None:
            self.problem.problem.setAttr('ModelSense', {'min': 1, 'max': -1}[value])

    def set_linear_coefficients(self, coefficients):
        if self.problem is not None:
            self.problem.update()
            for var, coeff in coefficients.items():
                var._internal_variable.setAttr("Obj", coeff)
            self._expression_expired = True
            self.problem.update()
        else:
            raise Exception("Can't change coefficients if objective is not associated with a model.")

    def get_linear_coefficients(self, variables):
        if self.problem is not None:
            self.problem.update()
            obj = self.problem.problem.getObjective()
            coefs = {obj.getVar(i).getAttr("varName"): obj.getCoeff(i) for i in range(obj.size())}
            return {v: coefs.get(v.name, 0) for v in variables}
        else:
            raise("Can't get coefficients from solver if objective is not in a model")

    def _get_expression(self):
        if self.problem is not None and self._expression_expired and len(self.problem._variables) > 0:
            grb_obj = self.problem.problem.getObjective()
            variables = self.problem._variables
            if self.problem.problem.IsQP:
                quadratic_expression = symbolics.add(
                    [symbolics.Real(grb_obj.getCoeff(i)) *
                     variables[grb_obj.getVar1(i).VarName] *
                     variables[grb_obj.getVar2(i).VarName]
                     for i in range(grb_obj.size())])
                linear_objective = grb_obj.getLinExpr()
            else:
                quadratic_expression = symbolics.Real(0.0)
                linear_objective = grb_obj
            linear_expression = symbolics.add(
                [symbolics.Real(linear_objective.getCoeff(i)) *
                 variables[linear_objective.getVar(i).VarName]
                 for i in range(linear_objective.size())]
            )
            self._expression = (linear_expression + quadratic_expression +
                                getattr(self.problem, "_objective_offset", 0))
            self._expression_expired = False
        return self._expression


@six.add_metaclass(inheritdocstring)
class Configuration(interface.MathematicalProgrammingConfiguration):
    def __init__(self, lp_method='primal', qp_method='primal', presolve=False,
                 verbosity=0, timeout=None, *args, **kwargs):
        super(Configuration, self).__init__(*args, **kwargs)
        self.verbosity = verbosity
        self.lp_method = lp_method
        self.qp_method = qp_method
        self.presolve = presolve
        self.timeout = timeout
        if "tolerances" in kwargs:
            for key, val in six.iteritems(kwargs["tolerances"]):
                try:
                    setattr(self.tolerances, key, val)
                except AttributeError:
                    pass

    @property
    def lp_method(self):
        return _REVERSE_LP_METHODS[self.problem.problem.params.Method]

    @lp_method.setter
    def lp_method(self, value):
        if value not in _LP_METHODS:
            raise ValueError("Invalid LP method. Please choose among: " + str(list(_LP_METHODS)))
        if getattr(self, "problem", None) is not None:
            self.problem.problem.params.Method = _LP_METHODS[value]

    @property
    def qp_method(self):
        return _REVERSE_QP_METHODS[self.problem.problem.params.Method]

    @qp_method.setter
    def qp_method(self, value):
        if value not in _QP_METHODS:
            raise ValueError("Invalid LP method. Please choose among: " + str(list(_QP_METHODS)))
        if getattr(self, "problem", None) is not None:
            self.problem.problem.params.Method = _QP_METHODS[value]

    @property
    def presolve(self):
        return self._presolve

    @presolve.setter
    def presolve(self, value):
        if getattr(self, "problem", None) is not None:
            if value is True:
                self.problem.problem.params.Presolve = 2
            elif value is False:
                self.problem.problem.params.Presolve = 0
            else:
                self.problem.problem.params.Presolve = -1
        self._presolve = value

    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value):
        if self.problem is not None:
            if value == 0:
                self.problem.problem.params.OutputFlag = 0
            elif 0 < value <= 3:
                self.problem.problem.params.OutputFlag = 1
        self._verbosity = value

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        if self.problem is not None:
            if value is None:
                self.problem.problem.params.TimeLimit = gurobipy.GRB.INFINITY
            else:
                self.problem.problem.params.TimeLimit = value
        self._timeout = value

    def __getstate__(self):
        return {'presolve': self.presolve, 'verbosity': self.verbosity, 'timeout': self.timeout}

    def __setstate__(self, state):
        self.__init__()
        for key, val in six.iteritems(state):
            setattr(self, key, val)

    def _get_feasibility(self):
        return getattr(self.problem.problem.params, "FeasibilityTol")

    def _set_feasibility(self, value):
        return setattr(self.problem.problem.params, "FeasibilityTol", value)

    def _get_optimality(self):
        return getattr(self.problem.problem.params, "OptimalityTol")

    def _set_optimality(self, value):
        return setattr(self.problem.problem.params, "OptimalityTol", value)

    def _get_integrality(self):
        return getattr(self.problem.problem.params, "IntFeasTol")

    def _set_integrality(self, value):
        return setattr(self.problem.problem.params, "IntFeasTol", value)

    def _tolerance_functions(self):
        return {
            "feasibility": (self._get_feasibility, self._set_feasibility),
            "optimality": (self._get_optimality, self._set_optimality),
            "integrality": (self._get_integrality, self._set_integrality)
        }


class Model(interface.Model):

    def _initialize_problem(self):
        self.problem = gurobipy.Model()
        self.problem.params.OutputFlag = 0
        if self.name is not None:
            self.problem.setAttr('ModelName', self.name)
            self.problem.update()

    def _initialize_model_from_problem(self, problem):
        if isinstance(problem, gurobipy.Model):
            self.problem = problem
            variables = []
            for gurobi_variable in self.problem.getVars():
                variables.append(Variable(
                    gurobi_variable.getAttr("VarName"),
                    lb=gurobi_variable.getAttr("lb"),
                    ub=gurobi_variable.getAttr("ub"),
                    problem=self,
                    type=_GUROBI_VTYPE_TO_VTYPE[gurobi_variable.getAttr("vType")]
                ))
            super(Model, self)._add_variables(variables)
            constraints = []
            for gurobi_constraint in self.problem.getConstrs():
                sense = gurobi_constraint.Sense
                name = gurobi_constraint.getAttr("ConstrName")
                rhs = gurobi_constraint.RHS
                row = self.problem.getRow(gurobi_constraint)
                lhs = symbolics.add(
                    [symbolics.Real(row.getCoeff(i)) * self.variables[row.getVar(i).VarName] for i in
                     range(row.size())]
                )

                if sense == '=':
                    constraint = Constraint(lhs, name=name, lb=rhs, ub=rhs, problem=self)
                elif sense == '>':
                    constraint = Constraint(lhs, name=name, lb=rhs, ub=None, problem=self)
                elif sense == '<':
                    constraint = Constraint(lhs, name=name, lb=None, ub=rhs, problem=self)
                else:
                    raise ValueError('{} is not a valid sense'.format(sense))
                constraints.append(constraint)
            super(Model, self)._add_constraints(constraints, sloppy=True)

            gurobi_objective = self.problem.getObjective()
            if self.problem.IsQP:
                quadratic_expression = symbolics.add(
                    [symbolics.Real(gurobi_objective.getCoeff(i)) *
                     self.variables[gurobi_objective.getVar1(i).VarName] *
                     self.variables[gurobi_objective.getVar2(i).VarName]
                     for i in range(gurobi_objective.size())])
                linear_objective = gurobi_objective.getLinExpr()
            else:
                quadratic_expression = symbolics.Real(0.0)
                linear_objective = gurobi_objective
            linear_expression = symbolics.add(
                [symbolics.Real(linear_objective.getCoeff(i)) *
                 self.variables[linear_objective.getVar(i).VarName]
                 for i in range(linear_objective.size())]
            )

            self._objective = Objective(
                quadratic_expression + linear_expression,
                problem=self,
                direction={1: 'min', -1: 'max'}[self.problem.getAttr('ModelSense')]
            )
        else:
            raise TypeError("Provided problem is not a valid Gurobi model.")

    @classmethod
    def from_lp(cls, lp_form):
        with TemporaryFilename(suffix=".lp", content=lp_form) as tmp_file_name:
            problem = gurobipy.read(tmp_file_name)
        model = cls(problem=problem)
        return model

    def __getstate__(self):
        self.update()
        with TemporaryFilename(suffix=".lp") as tmp_file_name:
            self.problem.write(tmp_file_name)
            with open(tmp_file_name) as tmp_file:
                lp_form = tmp_file.read()
        repr_dict = {'lp': lp_form, 'status': self.status, 'config': self.configuration}
        return repr_dict

    def __setstate__(self, repr_dict):
        with TemporaryFilename(suffix=".lp", content=repr_dict["lp"]) as tmp_file_name:
            problem = gurobipy.read(tmp_file_name)
        # if repr_dict['status'] == 'optimal':  # TODO: uncomment this
        #     # turn off logging completely, get's configured later
        #     problem.set_error_stream(None)
        #     problem.set_warning_stream(None)
        #     problem.set_log_stream(None)
        #     problem.set_results_stream(None)
        #     problem.solve()  # since the start is an optimal solution, nothing will happen here
        self.__init__(problem=problem)
        self.configuration = Configuration.clone(repr_dict['config'], problem=self)  # TODO: make configuration work

    def to_lp(self):
        self.problem.update()
        with TemporaryFilename(suffix=".lp") as tmp_file_name:
            self.problem.write(tmp_file_name)
            with open(tmp_file_name) as tmp_file:
                lp_form = tmp_file.read()
        return lp_form

    @property
    def objective(self):
        return self._objective

    @objective.setter
    def objective(self, value):
        super(Model, self.__class__).objective.fset(self, value)
        expression = self._objective._expression

        offset, linear_coefficients, quadratic_coefficients = parse_optimization_expression(
            value, quadratic=True, expression=expression
        )
        # self.problem.setAttr("ObjCon", offset) # Does not seem to work
        self._objective_offset = offset
        grb_terms = []
        for var, coef in linear_coefficients.items():
            coef = float(coef)
            var = self.problem.getVarByName(var.name)
            grb_terms.append(coef * var)
        for key, coef in quadratic_coefficients.items():
            coef = float(coef)
            if len(key) == 1:
                var = six.next(iter(key))
                var = self.problem.getVarByName(var.name)
                grb_terms.append(coef * var * var)
            else:
                var1, var2 = key
                var1 = self.problem.getVarByName(var1.name)
                var2 = self.problem.getVarByName(var2.name)
                grb_terms.append(coef * var1 * var2)

        grb_expression = gurobipy.quicksum(grb_terms)

        self.problem.setObjective(grb_expression,
                                  {'min': gurobipy.GRB.MINIMIZE, 'max': gurobipy.GRB.MAXIMIZE}[value.direction])
        value.problem = self

    def update(self):
        super(Model, self).update(callback=self.problem.update)

    def _optimize(self):
        if self.status != interface.OPTIMAL:
            self.problem.reset()
        self.problem.update()
        self.problem.optimize()
        status = _GUROBI_STATUS_TO_STATUS[self.problem.getAttr("Status")]
        return status

    def _add_variables(self, variables):
        super(Model, self)._add_variables(variables)
        for variable in variables:
            if variable.lb is None:
                lb = -gurobipy.GRB.INFINITY
            else:
                lb = variable.lb
            if variable.ub is None:
                ub = gurobipy.GRB.INFINITY
            else:
                ub = variable.ub
            self.problem.addVar(
                name=variable.name,
                lb=lb,
                ub=ub,
                vtype=_VTYPE_TO_GUROBI_VTYPE[variable.type]
            )

    def _remove_variables(self, variables):
        # Not calling parent method to avoid expensive variable removal from sympy expressions
        self.objective._expression = self.objective.expression.xreplace({var: 0 for var in variables})
        for variable in variables:
            internal_variable = variable._internal_variable
            del self._variables_to_constraints_mapping[variable.name]
            variable.problem = None
            del self._variables[variable.name]
            self.problem.remove(internal_variable)

    def _add_constraints(self, constraints, sloppy=False):
        super(Model, self)._add_constraints(constraints, sloppy=sloppy)
        for constraint in constraints:
            if constraint.lb is None and constraint.ub is None:
                raise ValueError("optlang does not support free constraints in the gurobi interface.")
            self.problem.update()
            constraint._problem = None
            if constraint.is_Linear:
                offset, coef_dict, _ = parse_optimization_expression(constraint, linear=True)

                lhs = gurobipy.quicksum([float(coef) * var._internal_variable for var, coef in coef_dict.items()])
                sense, rhs, range_value = _constraint_lb_and_ub_to_gurobi_sense_rhs_and_range_value(constraint.lb,
                                                                                                    constraint.ub)

                if range_value != 0:
                    aux_var = self.problem.addVar(name=constraint.name + '_aux', lb=0, ub=range_value)
                    self.problem.update()
                    lhs = lhs - aux_var

                self.problem.addConstr(lhs, sense, rhs, name=constraint.name)
            else:
                raise ValueError(
                    "GUROBI currently only supports linear constraints. %s is not linear." % self)
                # self.problem.addQConstr(lhs, sense, rhs)
            constraint.problem = self
        self.problem.update()

    def _remove_constraints(self, constraints):
        self.problem.update()
        internal_constraints = [constraint._internal_constraint for constraint in constraints]
        super(Model, self)._remove_constraints(constraints)
        for internal_constraint in internal_constraints:
            self.problem.remove(internal_constraint)

    @property
    def is_integer(self):
        self.problem.update()
        return self.problem.NumIntVars > 0


if __name__ == '__main__':
    x = Variable('x', lb=0, ub=10)
    y = Variable('y', lb=0, ub=10)
    z = Variable('z', lb=-100, ub=99.)
    constr = Constraint(0.3 * x + 0.4 * y + 66. * z, lb=-100, ub=0., name='test')

    solver = Model(problem=gurobipy.read("tests/data/model.lp"))

    solver.add(z)
    solver.add(constr)
    print(solver)
    print(solver.optimize())
    print(solver.status)



# Copyright 2017 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Interface for the GNU Linear Programming Kit (GLPK)

GLPK is an open source LP solver, with MILP capabilities. This interface exposes its GLPK's exact solver.
To use GLPK you need to install the 'swiglpk' python package (with pip or from http://github.com/biosustain/swiglpk)
and make sure that 'import swiglpk' runs without error.
"""

import logging

import six

from optlang.util import inheritdocstring
from optlang import interface
from optlang import glpk_interface
from optlang.glpk_interface import _GLPK_STATUS_TO_STATUS

log = logging.getLogger(__name__)

from swiglpk import glp_exact, glp_create_prob, glp_get_status, \
    GLP_SF_AUTO, GLP_ETMLIM, glp_adv_basis, glp_read_lp, glp_scale_prob


@six.add_metaclass(inheritdocstring)
class Variable(glpk_interface.Variable):
    def __init__(self, name, index=None, type="continuous", **kwargs):
        if type in ("integer", "binary"):
            raise ValueError("The GLPK exact solver does not support integer and mixed integer problems")
        super(Variable, self).__init__(name, index, type=type, **kwargs)

    @glpk_interface.Variable.type.setter
    def type(self, value):
        if value in ("integer", "binary"):
            raise ValueError("The GLPK exact solver does not support integer and mixed integer problems")
        super(Variable, Variable).type.fset(self, value)


@six.add_metaclass(inheritdocstring)
class Constraint(glpk_interface.Constraint):
    pass


@six.add_metaclass(inheritdocstring)
class Objective(glpk_interface.Objective):
    pass


@six.add_metaclass(inheritdocstring)
class Configuration(glpk_interface.Configuration):
    pass


@six.add_metaclass(inheritdocstring)
class Model(glpk_interface.Model):
    def _run_glp_exact(self):
        return_value = glp_exact(self.problem, self.configuration._smcp)
        glpk_status = glp_get_status(self.problem)
        if return_value == 0:
            status = _GLPK_STATUS_TO_STATUS[glpk_status]
        elif return_value == GLP_ETMLIM:
            status = interface.TIME_LIMIT
        else:
            status = _GLPK_STATUS_TO_STATUS[glpk_status]
            if status == interface.UNDEFINED:
                log.debug("Status undefined. GLPK status code returned by glp_simplex was %d" % return_value)
        return status

    def _optimize(self):
        # Solving inexact first per GLPK manual
        #    Computations in exact arithmetic are very time consuming, so solving LP
        #    problem with the routine glp_exact from the very beginning is not a good
        #    idea. It is much better at first to find an optimal basis with the routine
        #    glp_simplex and only then to call glp_exact, in which case only a few
        #    simplex iterations need to be performed in exact arithmetic.
        status = super(Model, self)._optimize()
        if status != interface.OPTIMAL:
            return status
        else:
            status = self._run_glp_exact()

            if status == interface.UNDEFINED and self.configuration.presolve is True:
                # If presolve is on, status will be undefined if not optimal
                self.configuration.presolve = False
                status = self._run_glp_exact()
                self.configuration.presolve = True
            return status


if __name__ == '__main__':
    import pickle

    x1 = Variable('x1', lb=0)
    x2 = Variable('x2', lb=0)
    x3 = Variable('x3', lb=0, ub=1, type='binary')
    c1 = Constraint(x1 + x2 + x3, lb=-100, ub=100, name='c1')
    c2 = Constraint(10 * x1 + 4 * x2 + 5 * x3, ub=600, name='c2')
    c3 = Constraint(2 * x1 + 2 * x2 + 6 * x3, ub=300, name='c3')
    obj = Objective(10 * x1 + 6 * x2 + 4 * x3, direction='max')
    model = Model(name='Simple model')
    model.objective = obj
    model.add([c1, c2, c3])
    model.configuration.verbosity = 3
    status = model.optimize()
    print("status:", model.status)
    print("objective value:", model.objective.value)

    for var_name, var in model.variables.items():
        print(var_name, "=", var.primal)

    print(model)

    problem = glp_create_prob()
    glp_read_lp(problem, None, "tests/data/model.lp")

    solver = Model(problem=problem)
    print(solver.optimize())
    print(solver.objective)

    import time

    t1 = time.time()
    print("pickling")
    pickle_string = pickle.dumps(solver)
    resurrected_solver = pickle.loads(pickle_string)
    t2 = time.time()
    print("Execution time: %s" % (t2 - t1))

    resurrected_solver.optimize()
    print("Halelujah!", resurrected_solver.objective.value)



# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import unittest

import nose
import optlang
import optlang.glpk_interface as glpk
from optlang import interface
from optlang.util import glpk_read_cplex

TESTMODELPATH = os.path.join(os.path.dirname(__file__), 'data/model.lp')

try:
    import optlang.cplex_interface as cplex


    class ChangeSolverTestCase(unittest.TestCase):
        def setUp(self):
            self.model = optlang.glpk_interface.Model(problem=glpk_read_cplex(TESTMODELPATH))

        def test_clone_constraint_interface_to_glpk(self):
            x1 = interface.Variable('x1', lb=0)
            x2 = interface.Variable('x2', lb=0)
            x3 = interface.Variable('x3', lb=0)
            c1 = interface.Constraint(x1 + x2 + x3, ub=100)
            glpk_c1 = glpk.Constraint.clone(c1)
            print(glpk_c1)
            self.assertIs(glpk_c1.__class__, glpk.Constraint)
            for variable in glpk_c1.variables:
                self.assertIs(variable.__class__, glpk.Variable)

        def test_clone_constraint_glpk_to_cplex(self):
            x1 = glpk.Variable('x1', lb=0)
            x2 = glpk.Variable('x2', lb=0)
            x3 = glpk.Variable('x3', lb=0)
            c1 = glpk.Constraint(x1 + x2 + x3, ub=100)
            cplex_c1 = cplex.Constraint.clone(c1)
            print(cplex_c1)
            self.assertIs(cplex_c1.__class__, cplex.Constraint)
            for variable in cplex_c1.variables:
                self.assertIs(variable.__class__, cplex.Variable)

        def test_clone_objective_glpk_to_cplex(self):
            x1 = glpk.Variable('x1', lb=0)
            x2 = glpk.Variable('x2', lb=0)
            x3 = glpk.Variable('x3', lb=0)
            obj = glpk.Objective(x1 + x2 + x3)
            cplex_obj = cplex.Objective.clone(obj)
            self.assertIs(cplex_obj.__class__, cplex.Objective)
            for variable in cplex_obj.variables:
                self.assertIs(variable.__class__, cplex.Variable)

        def test_clone_to_cplex(self):
            cplex_model = cplex.Model.clone(self.model)
            self.assertEqual(cplex_model.__class__, cplex.Model)
            for variable in cplex_model.variables.values():
                self.assertIs(variable.__class__, cplex.Variable)
            for constraint in cplex_model.constraints:
                self.assertIs(constraint.__class__, cplex.Constraint)

        def test_clone_to_glpk(self):
            cplex_model = cplex.Model.clone(self.model)
            glpk_model = glpk.Model.clone(cplex_model)
            self.assertEqual(glpk_model.__class__, glpk.Model)
            for variable in glpk_model.variables.values():
                self.assertIs(variable.__class__, glpk.Variable)
            for constraint in glpk_model.constraints:
                self.assertIs(constraint.__class__, glpk.Constraint)

except ImportError as e:

    if str(e).find('cplex') >= 0:
        class TestMissingDependency(unittest.TestCase):

            @unittest.skip('Missing dependency - ' + str(e))
            def test_fail(self):
                pass
    else:
        raise

if __name__ == '__main__':
    nose.runmodule()



# Copyright (c) 2013 Novo Nordisk Foundation Center for Biosustainability, DTU.
# See LICENSE for details.

import glob
import os
import pickle
import tarfile
import tempfile
import unittest
from functools import partial

import nose
import six

if six.PY3:
    nose.SkipTest('Skipping because py3')
elif os.getenv('CI', 'false') == 'true':
    nose.SkipTest('Skipping because solving problems takes very long.')
else:
    try:
        import cplex

        from optlang.cplex_interface import Model

        with open(os.path.join(os.path.dirname(__file__), 'data/the_final_netlib_results.pcl'), 'rb') as fhandle:
            THE_FINAL_NETLIB_RESULTS = pickle.load(fhandle)


        class TestClass(object):
            """docstring for TestClass"""

            def __init__(self, arg):
                super(TestClass, self).__init__()
                self.arg = arg


        # noinspection PyShadowingNames
        def read_netlib_sif_cplex(fhandle):
            tmp_file = tempfile.mktemp(suffix='.sif')
            with open(tmp_file, 'w') as tmp_handle:
                content = ''.join([str(s) for s in fhandle if str(s).strip()])
                tmp_handle.write(content)
                fhandle.close()
            problem = cplex.Cplex()
            problem.read(tmp_file)
            # glp_read_mps(problem, GLP_MPS_FILE, None, tmp_file)
            return problem


        def check_dimensions(problem, model):
            """
            Tests that the glpk problem and the interface model have the same
            number of rows (constraints) and columns (variables).
            """
            assert problem.variables.get_num() == len(model.variables)


        def check_objval(problem, model_objval):
            """
            Check that ...
            """
            if problem.solution.get_status() == cplex.Cplex().solution.status.optimal:
                problem_objval = problem.solution.get_objective_value()
            else:
                problem_objval = None
            nose.tools.assert_almost_equal(problem_objval, model_objval, places=4)


        def check_objval_against_the_final_netlib_results(netlib_id, model_objval):
            relative_error = abs(1 - (model_objval / float(THE_FINAL_NETLIB_RESULTS[netlib_id]['Objvalue'])))
            print(relative_error)
            nose.tools.assert_true(relative_error < 0.01)
            # nose.tools.assert_almost_equal(model_objval, float(THE_FINAL_NETLIB_RESULTS[netlib_id]['Objvalue']), places=4)


        def test_netlib(netlib_tar_path=os.path.join(os.path.dirname(__file__), 'data/netlib_lp_problems.tar.gz')):
            """
            Test netlib with glpk interface
            """
            tar = tarfile.open(netlib_tar_path)
            model_paths_in_tar = glob.fnmatch.filter(tar.getnames(), '*.SIF')

            for model_path_in_tar in model_paths_in_tar:
                print(model_path_in_tar)
                netlib_id = os.path.basename(model_path_in_tar).replace('.SIF', '')
                # TODO: get the following problems to work
                # E226 seems to be a MPS related problem, see http://lists.gnu.org/archive/html/bug-glpk/2003-01/msg00003.html
                if netlib_id in ('AGG', 'E226', 'SCSD6', 'BLEND', 'DFL001', 'FORPLAN', 'GFRD-PNC', 'SIERRA'):
                    # def test_skip(netlib_id):
                    # raise SkipTest('Skipping netlib problem %s ...' % netlib_id)
                    # test_skip(netlib_id)
                    # class TestWeirdNetlibProblems(unittest.TestCase):

                    # @unittest.skip('Skipping netlib problem')
                    # def test_fail():
                    # pass
                    continue
                # TODO: For now, test only models that are covered by the final netlib results
                else:
                    if netlib_id not in THE_FINAL_NETLIB_RESULTS.keys():
                        continue
                    fhandle = tar.extractfile(model_path_in_tar)
                    problem = read_netlib_sif_cplex(fhandle)
                    model = Model(problem=problem)
                    model.configuration.presolve = True
                    model.configuration.verbosity = 3
                    func = partial(check_dimensions, problem, model)
                    func.description = "test_netlib_check_dimensions_%s (%s)" % (
                        netlib_id, os.path.basename(str(__file__)))
                    yield func

                    model.optimize()
                    if model.status == 'optimal':
                        model_objval = model.objective.value
                    else:
                        raise Exception('No optimal solution found for netlib model %s' % netlib_id)

                    func = partial(check_objval, problem, model_objval)
                    func.description = "test_netlib_check_objective_value_%s (%s)" % (
                        netlib_id, os.path.basename(str(__file__)))
                    yield func

                    func = partial(check_objval_against_the_final_netlib_results, netlib_id, model_objval)
                    func.description = "test_netlib_check_objective_value__against_the_final_netlib_results_%s (%s)" % (
                        netlib_id, os.path.basename(str(__file__)))
                    yield func

                    if os.getenv('CI', 'false') != 'true':
                        # check that a cloned model also gives the correct result
                        model = Model.clone(model, use_json=False, use_lp=False)
                        model.optimize()
                        if model.status == 'optimal':
                            model_objval = model.objective.value
                        else:
                            raise Exception('No optimal solution found for netlib model %s' % netlib_id)

                        func = partial(check_objval_against_the_final_netlib_results, netlib_id, model_objval)
                        func.description = "test_netlib_check_objective_value__against_the_final_netlib_results_after_cloning_%s (%s)" % (
                            netlib_id, os.path.basename(str(__file__)))
                        yield func

    except ImportError as e:

        if str(e).find('cplex') >= 0:
            class TestMissingDependency(unittest.TestCase):

                @unittest.skip('Missing dependency - ' + str(e))
                def test_fail(self):
                    pass
        else:
            raise

if __name__ == '__main__':
    # tar = tarfile.open('data/netlib_lp_problems.tar.gz')
    # fhandle = tar.extractfile('./netlib/SEBA.SIF')
    # problem = read_netlib_sif_cplex(fhandle)
    # model = Model(problem=problem)
    # status = model.optimize()
    # print(status)
    # print(model.objective.value)
    # print(model.constraints['VILLKOR7'])
    # # print(model)
    # model = Model.clone(model, use_lp=False, use_json=False)
    # status = model.optimize()
    # print(status)
    # print(model.constraints['VILLKOR7'])
    # print(model.objective.value)
    nose.runmodule()



# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Solver interface for the IBM ILOG CPLEX Optimization Studio solver.

Wraps the cplex solver by subclassing and extending :class:`Model`,
:class:`Variable`, and :class:`Constraint` from :mod:`interface`.

To use this interface, install the cplex solver and the bundled python interface.
Make sure that 'import cplex' runs without error.
"""
import logging
import sys

import six
import os
from six.moves import StringIO

import cplex
from cplex.exceptions import CplexSolverError

from optlang import interface
from optlang import symbolics
from optlang.util import inheritdocstring, TemporaryFilename
from optlang.expression_parsing import parse_optimization_expression
from optlang.exceptions import SolverError

log = logging.getLogger(__name__)

from optlang.symbolics import add, mul, One, Zero

_STATUS_MAP = {
    'MIP_abort_feasible': interface.ABORTED,
    'MIP_abort_infeasible': interface.ABORTED,
    'MIP_dettime_limit_feasible': interface.TIME_LIMIT,
    'MIP_dettime_limit_infeasible': interface.TIME_LIMIT,
    'MIP_feasible': interface.FEASIBLE,
    'MIP_feasible_relaxed_inf': interface.SPECIAL,
    'MIP_feasible_relaxed_quad': interface.SPECIAL,
    'MIP_feasible_relaxed_sum': interface.SPECIAL,
    'MIP_infeasible': interface.INFEASIBLE,
    'MIP_infeasible_or_unbounded': interface.INFEASIBLE_OR_UNBOUNDED,
    'MIP_optimal': interface.OPTIMAL,
    'MIP_optimal_infeasible': interface.SPECIAL,
    'MIP_optimal_relaxed_inf': interface.SPECIAL,
    'MIP_optimal_relaxed_sum': interface.SPECIAL,
    'MIP_time_limit_feasible': interface.TIME_LIMIT,
    'MIP_time_limit_infeasible': interface.TIME_LIMIT,
    'MIP_unbounded': interface.UNBOUNDED,
    'abort_dettime_limit': interface.ABORTED,
    'abort_dual_obj_limit': interface.ABORTED,
    'abort_iteration_limit': interface.ABORTED,
    'abort_obj_limit': interface.ABORTED,
    'abort_primal_obj_limit': interface.ABORTED,
    'abort_relaxed': interface.ABORTED,
    'abort_time_limit': interface.TIME_LIMIT,
    'abort_user': interface.ABORTED,
    'conflict_abort_contradiction': interface.SPECIAL,
    'conflict_abort_dettime_limit': interface.SPECIAL,
    'conflict_abort_iteration_limit': interface.SPECIAL,
    'conflict_abort_memory_limit': interface.SPECIAL,
    'conflict_abort_node_limit': interface.SPECIAL,
    'conflict_abort_obj_limit': interface.SPECIAL,
    'conflict_abort_time_limit': interface.SPECIAL,
    'conflict_abort_user': interface.SPECIAL,
    'conflict_feasible': interface.SPECIAL,
    'conflict_minimal': interface.SPECIAL,
    'fail_feasible': interface.SPECIAL,
    'fail_feasible_no_tree': interface.SPECIAL,
    'fail_infeasible': interface.SPECIAL,
    'fail_infeasible_no_tree': interface.SPECIAL,
    'feasible': interface.FEASIBLE,
    'feasible_relaxed_inf': interface.SPECIAL,
    'feasible_relaxed_quad': interface.SPECIAL,
    'feasible_relaxed_sum': interface.SPECIAL,
    'first_order': interface.SPECIAL,
    'infeasible': interface.INFEASIBLE,
    'infeasible_or_unbounded': interface.INFEASIBLE_OR_UNBOUNDED,
    'mem_limit_feasible': interface.MEMORY_LIMIT,
    'mem_limit_infeasible': interface.MEMORY_LIMIT,
    'node_limit_feasible': interface.NODE_LIMIT,
    'node_limit_infeasible': interface.NODE_LIMIT,
    'num_best': interface.NUMERIC,
    'optimal': interface.OPTIMAL,
    'optimal_face_unbounded': interface.SPECIAL,
    'optimal_infeasible': interface.INFEASIBLE,
    'optimal_populated': interface.SPECIAL,
    'optimal_populated_tolerance': interface.SPECIAL,
    'optimal_relaxed_inf': interface.SPECIAL,
    'optimal_relaxed_quad': interface.SPECIAL,
    'optimal_relaxed_sum': interface.SPECIAL,
    'optimal_tolerance': interface.OPTIMAL,
    'populate_solution_limit': interface.SPECIAL,
    'solution_limit': interface.SPECIAL,
    'unbounded': interface.UNBOUNDED,
    'relaxation_unbounded': interface.UNBOUNDED,
    'non-existing-status': 'Here for testing that missing statuses are handled.'
    # 102: interface.OPTIMAL # The same as cplex.Cplex.solution.status.optimal_tolerance
}

# Check if each status is supported by the current cplex version
_CPLEX_STATUS_TO_STATUS = {}
_solution = cplex.Cplex().solution
for status_name, optlang_status in _STATUS_MAP.items():
    cplex_status = getattr(_solution.status, status_name, None)
    if cplex_status is not None:
        _CPLEX_STATUS_TO_STATUS[cplex_status] = optlang_status

_LP_METHODS = ["auto", "primal", "dual", "network", "barrier", "sifting", "concurrent"]

_SOLUTION_TARGETS = ("auto", "convex", "local", "global")

_QP_METHODS = ("auto", "primal", "dual", "network", "barrier")

_CPLEX_VTYPE_TO_VTYPE = {'C': 'continuous', 'I': 'integer', 'B': 'binary'}

_VTYPE_TO_CPLEX_VTYPE = dict(
    [(val, key) for key, val in six.iteritems(_CPLEX_VTYPE_TO_VTYPE)]
)

_CPLEX_MIP_TYPES_TO_CONTINUOUS = {
    cplex.Cplex.problem_type.MILP: cplex.Cplex.problem_type.LP,
    cplex.Cplex.problem_type.MIQP: cplex.Cplex.problem_type.QP,
    cplex.Cplex.problem_type.MIQCP: cplex.Cplex.problem_type.QCP
}


def _constraint_lb_and_ub_to_cplex_sense_rhs_and_range_value(lb, ub):
    """Helper function used by Constraint and Model"""
    if lb is None and ub is None:
        raise Exception("Free constraint ...")
    elif lb is None:
        sense = 'L'
        rhs = float(ub)
        range_value = 0.
    elif ub is None:
        sense = 'G'
        rhs = float(lb)
        range_value = 0.
    elif lb == ub:
        sense = 'E'
        rhs = float(lb)
        range_value = 0.
    elif lb > ub:
        raise ValueError("Lower bound is larger than upper bound.")
    else:
        sense = 'R'
        rhs = float(lb)
        range_value = float(ub - lb)
    return sense, rhs, range_value


@six.add_metaclass(inheritdocstring)
class Variable(interface.Variable):
    def __init__(self, name, *args, **kwargs):
        super(Variable, self).__init__(name, **kwargs)

    @interface.Variable.type.setter
    def type(self, value):
        if self.problem is not None:
            try:
                cplex_kind = _VTYPE_TO_CPLEX_VTYPE[value]
            except KeyError:
                raise ValueError(
                    "CPLEX cannot handle variables of type '%s'. " % value +
                    "The following variable types are available: " +
                    ", ".join(_VTYPE_TO_CPLEX_VTYPE.keys())
                )
            self.problem.problem.variables.set_types(self.name, cplex_kind)
            if value == "continuous" and not self.problem.is_integer:
                cplex_type = self.problem.problem.get_problem_type()
                if cplex_type in _CPLEX_MIP_TYPES_TO_CONTINUOUS:
                    self.problem.problem.set_problem_type(_CPLEX_MIP_TYPES_TO_CONTINUOUS[cplex_type])
        super(Variable, Variable).type.fset(self, value)

    def _get_primal(self):
        try:
            return self.problem.problem.solution.get_values(self.name)
        except CplexSolverError as err:
            raise SolverError(str(err))

    @property
    def dual(self):
        if self.problem is None:
            return None
        if self.problem.is_integer:
            raise ValueError("Dual values are not well-defined for integer problems")
        try:
            return self.problem.problem.solution.get_reduced_costs(self.name)
        except CplexSolverError as err:
            raise SolverError(str(err))

    @interface.Variable.name.setter
    def name(self, value):
        old_name = getattr(self, "name", None)
        super(Variable, Variable).name.fset(self, value)
        if getattr(self, "problem", None) is not None:
            self.problem.problem.variables.set_names(old_name, value)


@six.add_metaclass(inheritdocstring)
class Constraint(interface.Constraint):
    _INDICATOR_CONSTRAINT_SUPPORT = True

    def __init__(self, expression, sloppy=False, *args, **kwargs):
        super(Constraint, self).__init__(expression, *args, sloppy=sloppy, **kwargs)

    def set_linear_coefficients(self, coefficients):
        if self.problem is not None:
            self.problem.update()
            triplets = [(self.name, var.name, float(coeff)) for var, coeff in six.iteritems(coefficients)]
            self.problem.problem.linear_constraints.set_coefficients(triplets)
        else:
            raise Exception("Can't change coefficients if constraint is not associated with a model.")

    def get_linear_coefficients(self, variables):
        if self.problem is not None:
            self.problem.update()
            coefs = self.problem.problem.linear_constraints.get_coefficients([(self.name, v.name) for v in variables])
            return {v: c for v, c in zip(variables, coefs)}
        else:
            raise Exception("Can't get coefficients from solver if constraint is not in a model")

    def _get_expression(self):
        if self.problem is not None:
            cplex_problem = self.problem.problem
            try:
                cplex_row = cplex_problem.linear_constraints.get_rows(self.name)
            except CplexSolverError as e:
                if 'CPLEX Error  1219:' not in str(e):
                    raise e
                else:
                    cplex_row = cplex_problem.indicator_constraints.get_linear_components(self.name)
            variables = self.problem._variables
            expression = add(
                [mul((symbolics.Real(cplex_row.val[i]), variables[ind])) for i, ind in
                 enumerate(cplex_row.ind)])
            self._expression = expression
        return self._expression

    @property
    def problem(self):
        return self._problem

    @problem.setter
    def problem(self, value):
        if value is None:
            # Update expression from solver instance one last time
            self._get_expression()
        self._problem = value

    @property
    def primal(self):
        if self.problem is None:
            return None
        try:
            # return self._round_primal_to_bounds(primal_from_solver)  # Test assertions fail
            return self.problem.problem.solution.get_activity_levels(self.name)
        except CplexSolverError as err:
            raise SolverError(str(err))

    @property
    def dual(self):
        if self.problem is None:
            return None
        if self.problem.is_integer:
            raise ValueError("Dual values are not well-defined for integer problems")
        try:
            return self.problem.problem.solution.get_dual_values(self.name)
        except CplexSolverError as err:
            raise SolverError(str(err))

    @interface.Constraint.name.setter
    def name(self, value):
        old_name = self.name
        super(Constraint, Constraint).name.fset(self, value)
        if getattr(self, 'problem', None) is not None:
            if self.indicator_variable is not None:
                raise NotImplementedError(
                    "Unfortunately, the CPLEX python bindings don't support changing an indicator constraint's name"
                )
            else:
                self.problem.problem.linear_constraints.set_names(old_name, value)

    @interface.Constraint.lb.setter
    def lb(self, value):
        self._check_valid_lower_bound(value)
        if getattr(self, 'problem', None) is not None:
            if self.indicator_variable is not None:
                raise NotImplementedError(
                    "Unfortunately, the CPLEX python bindings don't support changing an indicator constraint's bounds"
                )
            sense, rhs, range_value = _constraint_lb_and_ub_to_cplex_sense_rhs_and_range_value(value, self.ub)
            if self.is_Linear:
                self.problem.problem.linear_constraints.set_rhs(self.name, rhs)
                self.problem.problem.linear_constraints.set_senses(self.name, sense)
                self.problem.problem.linear_constraints.set_range_values(self.name, range_value)
        self._lb = value

    @interface.Constraint.ub.setter
    def ub(self, value):
        self._check_valid_upper_bound(value)
        if getattr(self, 'problem', None) is not None:
            if self.indicator_variable is not None:
                raise NotImplementedError(
                    "Unfortunately, the CPLEX python bindings don't support changing an indicator constraint's bounds"
                )
            sense, rhs, range_value = _constraint_lb_and_ub_to_cplex_sense_rhs_and_range_value(self.lb, value)
            if self.is_Linear:
                self.problem.problem.linear_constraints.set_rhs(self.name, rhs)
                self.problem.problem.linear_constraints.set_senses(self.name, sense)
                self.problem.problem.linear_constraints.set_range_values(self.name, range_value)
        self._ub = value

    def __iadd__(self, other):
        # if self.problem is not None:
        #     self.problem._add_to_constraint(self.index, other)
        if self.problem is not None:
            problem_reference = self.problem
            self.problem._remove_constraint(self)
            super(Constraint, self).__iadd__(other)
            problem_reference._add_constraint(self, sloppy=False)
        else:
            super(Constraint, self).__iadd__(other)
        return self


@six.add_metaclass(inheritdocstring)
class Objective(interface.Objective):
    def __init__(self, expression, sloppy=False, **kwargs):
        super(Objective, self).__init__(expression, sloppy=sloppy, **kwargs)
        self._expression_expired = False
        if not (sloppy or self.is_Linear or self.is_Quadratic):
            raise ValueError("Cplex only supports linear and quadratic objectives.")

    @property
    def value(self):
        if getattr(self, 'problem', None) is None:
            return None
        try:
            return self.problem.problem.solution.get_objective_value() + getattr(self.problem, "_objective_offset", 0)
        except CplexSolverError as err:
            raise SolverError(str(err))

    @interface.Objective.direction.setter
    def direction(self, value):
        if getattr(self, 'problem', None) is not None:
            self.problem.problem.objective.set_sense(
                {'min': self.problem.problem.objective.sense.minimize,
                 'max': self.problem.problem.objective.sense.maximize}[value])
        super(Objective, Objective).direction.__set__(self, value)

    def _get_expression(self):
        if self.problem is not None and self._expression_expired and len(self.problem._variables) > 0:
            cplex_problem = self.problem.problem
            coeffs = cplex_problem.objective.get_linear()
            expression = add([coeff * var for coeff, var in zip(coeffs, self.problem._variables) if coeff != 0.])
            if cplex_problem.objective.get_num_quadratic_nonzeros() > 0:
                expression += self.problem._get_quadratic_expression(cplex_problem.objective.get_quadratic())
            self._expression = expression + getattr(self.problem, "_objective_offset", 0)
            self._expression_expired = False
        return self._expression

    def set_linear_coefficients(self, coefficients):
        if self.problem is not None:
            self.problem.update()
            self.problem.problem.objective.set_linear([(variable.name, float(coefficient)) for variable, coefficient in coefficients.items()])
            self._expression_expired = True
        else:
            raise Exception("Can't change coefficients if objective is not associated with a model.")

    def get_linear_coefficients(self, variables):
        if self.problem is not None:
            self.problem.update()
            coefs = self.problem.problem.objective.get_linear([v.name for v in variables])
            return {v: c for v, c in zip(variables, coefs)}
        else:
            raise Exception("Can't get coefficients from solver if objective is not in a model")


@six.add_metaclass(inheritdocstring)
class Configuration(interface.MathematicalProgrammingConfiguration):
    def __init__(self, lp_method='primal', presolve="auto", verbosity=0, timeout=None,
                 solution_target="auto", qp_method="primal", *args, **kwargs):
        super(Configuration, self).__init__(*args, **kwargs)
        self.lp_method = lp_method
        self.presolve = presolve
        self.verbosity = verbosity
        self.timeout = timeout
        self.solution_target = solution_target
        self.qp_method = qp_method
        if "tolerances" in kwargs:
            for key, val in six.iteritems(kwargs["tolerances"]):
                try:
                    setattr(self.tolerances, key, val)
                except AttributeError:
                    pass

    @property
    def lp_method(self):
        """The algorithm used to solve LP problems."""
        lpmethod = self.problem.problem.parameters.lpmethod
        value = lpmethod.get()
        return lpmethod.values[value]

    @lp_method.setter
    def lp_method(self, lp_method):
        if lp_method not in _LP_METHODS:
            raise ValueError("LP Method %s is not valid (choose one of: %s)" % (lp_method, ", ".join(_LP_METHODS)))
        lp_method = getattr(self.problem.problem.parameters.lpmethod.values, lp_method)
        self.problem.problem.parameters.lpmethod.set(lp_method)

    def _set_presolve(self, value):
        if getattr(self, 'problem', None) is not None:
            presolve = self.problem.problem.parameters.preprocessing.presolve
            if value is True:
                presolve.set(presolve.values.on)
            elif value is False or value == "auto":
                presolve.set(presolve.values.off)
            else:
                raise ValueError(str(value) + " is not a valid presolve parameter. Must be True, False or 'auto'.")

    @property
    def presolve(self):
        return self._presolve

    @presolve.setter
    def presolve(self, value):
        self._set_presolve(value)
        self._presolve = value

    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value):

        class StreamHandler(StringIO):

            def __init__(self, logger, *args, **kwargs):
                StringIO.__init__(self, *args, **kwargs)
                self.logger = logger

        class ErrorStreamHandler(StreamHandler):

            def flush(self):
                self.logger.error(self.getvalue())

        class WarningStreamHandler(StreamHandler):

            def flush(self):
                self.logger.warning(self.getvalue())

        class LogStreamHandler(StreamHandler):

            def flush(self):
                self.logger.debug(self.getvalue())

        class ResultsStreamHandler(StreamHandler):

            def flush(self):
                self.logger.debug(self.getvalue())

        logger = logging.getLogger(__name__ + ".Model")  # TODO: Make the logger name specific to each solver instance
        logger.setLevel(logging.CRITICAL)
        error_stream_handler = ErrorStreamHandler(logger)
        warning_stream_handler = WarningStreamHandler(logger)
        log_stream_handler = LogStreamHandler(logger)
        results_stream_handler = ResultsStreamHandler(logger)
        if getattr(self, 'problem', None) is not None:
            problem = self.problem.problem
            if value == 0:
                problem.set_error_stream(error_stream_handler)
                problem.set_warning_stream(warning_stream_handler)
                problem.set_log_stream(None)
                problem.set_results_stream(None)
            elif value == 1:
                problem.set_error_stream(sys.stderr)
                problem.set_warning_stream(warning_stream_handler)
                problem.set_log_stream(log_stream_handler)
                problem.set_results_stream(results_stream_handler)
            elif value == 2:
                problem.set_error_stream(sys.stderr)
                problem.set_warning_stream(sys.stderr)
                problem.set_log_stream(log_stream_handler)
                problem.set_results_stream(results_stream_handler)
            elif value == 3:
                problem.set_error_stream(sys.stderr)
                problem.set_warning_stream(sys.stderr)
                problem.set_log_stream(sys.stdout)
                problem.set_results_stream(sys.stdout)
            else:
                raise ValueError(
                    "%s is not a valid verbosity level ranging between 0 and 3."
                    % value
                )
        self._verbosity = value

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        if getattr(self, 'problem', None) is not None:
            if value is None:
                self.problem.problem.parameters.timelimit.reset()
            else:
                self.problem.problem.parameters.timelimit.set(value)
        self._timeout = value

    def __getstate__(self):
        return {"presolve": self.presolve,
                "timeout": self.timeout,
                "verbosity": self.verbosity,
                "tolerances": {"feasibility": self.tolerances.feasibility,
                               "optimality": self.tolerances.optimality,
                               "integrality": self.tolerances.integrality}
                }

    def __setstate__(self, state):
        for key, val in six.iteritems(state):
            if key != "tolerances":
                setattr(self, key, val)

    @property
    def solution_target(self):
        """Change whether the QP solver will try to find a globally optimal solution or a local optimum.
        This will only"""
        if self.problem is not None:
            params = self.problem.problem.parameters
            try:
                solution_target = params.optimalitytarget
            except AttributeError:  # pragma: no cover
                solution_target = params.solutiontarget  # Compatibility with cplex < 12.6.3
            return _SOLUTION_TARGETS[solution_target.get()]
        else:
            return None

    @solution_target.setter
    def solution_target(self, value):
        if self.problem is not None:
            params = self.problem.problem.parameters
            try:
                solution_target = params.optimalitytarget
            except AttributeError:  # pragma: no cover
                solution_target = params.solutiontarget  # Compatibility with cplex < 12.6.3
            if value is None:
                solution_target.reset()
            else:
                try:
                    target = _SOLUTION_TARGETS.index(value)
                except ValueError:
                    raise ValueError(
                        "%s is not a valid solution target. Choose between %s" % (value, str(_SOLUTION_TARGETS)))
                solution_target.set(target)
        self._solution_target = self.solution_target

    @property
    def qp_method(self):
        """Change the algorithm used to optimize QP problems."""
        value = self.problem.problem.parameters.qpmethod.get()
        return self.problem.problem.parameters.qpmethod.values[value]

    @qp_method.setter
    def qp_method(self, value):
        if value not in _QP_METHODS:
            raise ValueError("%s is not a valid qp_method. Choose between %s" % (value, str(_QP_METHODS)))
        method = getattr(self.problem.problem.parameters.qpmethod.values, value)
        self.problem.problem.parameters.qpmethod.set(method)
        self._qp_method = value

    def _tolerance_functions(self):
        return {
            "feasibility": (
                self.problem.problem.parameters.simplex.tolerances.feasibility.get,
                self.problem.problem.parameters.simplex.tolerances.feasibility.set
            ),
            "optimality": (
                self.problem.problem.parameters.simplex.tolerances.optimality.get,
                self.problem.problem.parameters.simplex.tolerances.optimality.set
            ),
            "integrality": (
                self.problem.problem.parameters.mip.tolerances.integrality.get,
                self.problem.problem.parameters.mip.tolerances.integrality.set
            )
        }


@six.add_metaclass(inheritdocstring)
class Model(interface.Model):
    def _initialize_problem(self):
        self.problem = cplex.Cplex()

    def _initialize_model_from_problem(self, problem):
        if isinstance(problem, cplex.Cplex):
            self.problem = problem
            zipped_var_args = zip(self.problem.variables.get_names(),
                                  self.problem.variables.get_lower_bounds(),
                                  self.problem.variables.get_upper_bounds(),
                                  # self.problem.variables.get_types(), # TODO uncomment when cplex is fixed
                                  )
            for name, lb, ub in zipped_var_args:
                var = Variable(name, lb=lb, ub=ub, problem=self)  # Type should also be in there
                super(Model, self)._add_variables([var])  # This avoids adding the variable to the glpk problem
            zipped_constr_args = zip(self.problem.linear_constraints.get_names(),
                                     self.problem.linear_constraints.get_rows(),
                                     self.problem.linear_constraints.get_senses(),
                                     self.problem.linear_constraints.get_rhs()

                                     )
            variables = self._variables
            for name, row, sense, rhs in zipped_constr_args:
                constraint_variables = [variables[i - 1] for i in row.ind]

                # Since constraint expressions are lazily retrieved from the solver they don't have to be built here
                # lhs = _unevaluated_Add(*[val * variables[i - 1] for i, val in zip(row.ind, row.val)])
                lhs = symbolics.Integer(0)
                if sense == 'E':
                    constr = Constraint(lhs, lb=rhs, ub=rhs, name=name, problem=self)
                elif sense == 'G':
                    constr = Constraint(lhs, lb=rhs, name=name, problem=self)
                elif sense == 'L':
                    constr = Constraint(lhs, ub=rhs, name=name, problem=self)
                elif sense == 'R':
                    range_val = self.problem.linear_constraints.get_range_values(name)
                    if range_val > 0:
                        constr = Constraint(lhs, lb=rhs, ub=rhs + range_val, name=name, problem=self)
                    else:
                        constr = Constraint(lhs, lb=rhs + range_val, ub=rhs, name=name, problem=self)
                else:  # pragma: no cover
                    raise Exception('%s is not a recognized constraint sense.' % sense)

                for variable in constraint_variables:
                    try:
                        self._variables_to_constraints_mapping[variable.name].add(name)
                    except KeyError:
                        self._variables_to_constraints_mapping[variable.name] = set([name])

                super(Model, self)._add_constraints(
                    [constr],
                    sloppy=True
                )
            try:
                objective_name = self.problem.objective.get_name()
            except CplexSolverError as e:
                if 'CPLEX Error  1219:' not in str(e):
                    raise e
            else:
                linear_expression = add(
                    [mul(symbolics.Real(coeff), variables[index]) for index, coeff in enumerate(self.problem.objective.get_linear()) if coeff != 0.]
                )
                try:
                    quadratic = self.problem.objective.get_quadratic()
                except IndexError:
                    quadratic_expression = Zero
                else:
                    quadratic_expression = self._get_quadratic_expression(quadratic)

                self._objective = Objective(
                    linear_expression + quadratic_expression,
                    problem=self,
                    direction=
                    {self.problem.objective.sense.minimize: 'min', self.problem.objective.sense.maximize: 'max'}[
                        self.problem.objective.get_sense()],
                    name=objective_name
                )
        else:
            raise TypeError("Provided problem is not a valid CPLEX model.")

    @classmethod
    def from_lp(cls, lp_form):
        problem = cplex.Cplex()
        with TemporaryFilename(suffix=".lp", content=lp_form) as tmp_file_name:
            problem.read(tmp_file_name)
        model = cls(problem=problem)
        return model

    def __getstate__(self):
        self.update()
        with TemporaryFilename(suffix=".sav") as tmp_file_name:
            self.problem.write(tmp_file_name)
            with open(tmp_file_name, "rb") as tmp_file:
                cplex_binary = tmp_file.read()
        repr_dict = {'cplex_binary': cplex_binary, 'status': self.status, 'config': self.configuration}
        return repr_dict

    def __setstate__(self, repr_dict):
        with TemporaryFilename(suffix=".sav") as tmp_file_name:
            with open(tmp_file_name, "wb") as tmp_file:
                tmp_file.write(repr_dict["cplex_binary"])
            problem = cplex.Cplex()
            # turn off logging completely, get's configured later
            problem.set_error_stream(None)
            problem.set_warning_stream(None)
            problem.set_log_stream(None)
            problem.set_results_stream(None)
            problem.read(tmp_file_name)
        if repr_dict['status'] == 'optimal':
            problem.solve()  # since the start is an optimal solution, nothing will happen here
        self.__init__(problem=problem)
        self.configuration = Configuration.clone(repr_dict['config'], problem=self)

    @property
    def objective(self):
        return self._objective

    @objective.setter
    def objective(self, value):
        value.problem = None
        if self._objective is not None:  # Reset previous objective
            variables = self.objective.variables
            if len(variables) > 0:
                name_list = [var.name for var in variables]
                index_dict = {n: i for n, i in zip(name_list, self._get_variable_indices(name_list))}
                self.problem.objective.set_linear([(index_dict[variable.name], 0.) for variable in variables])
            if self.problem.objective.get_num_quadratic_variables() > 0:
                self.problem.objective.set_quadratic([0. for _ in range(self.problem.variables.get_num())])
        super(Model, self.__class__).objective.fset(self, value)
        self.update()
        expression = self._objective._expression
        offset, linear_coefficients, quadratic_coeffients = parse_optimization_expression(value, quadratic=True, expression=expression)
        # self.problem.objective.set_offset(float(offset)) # Not available prior to 12.6.2
        self._objective_offset = offset
        if linear_coefficients:
            name_list = [var.name for var in linear_coefficients]
            index_dict = {n: i for n, i in zip(name_list, self._get_variable_indices(name_list))}
            self.problem.objective.set_linear([index_dict[var.name], float(coef)] for var, coef in linear_coefficients.items())

        for key, coef in quadratic_coeffients.items():
            if len(key) == 1:
                var = six.next(iter(key))
                self.problem.objective.set_quadratic_coefficients(var.name, var.name, float(coef) * 2)
            else:
                var1, var2 = key
                self.problem.objective.set_quadratic_coefficients(var1.name, var2.name, float(coef))


        self._set_objective_direction(value.direction)
        self.problem.objective.set_name(value.name)
        value.problem = self

    def _set_objective_direction(self, direction):
        self.problem.objective.set_sense(
            {'min': self.problem.objective.sense.minimize, 'max': self.problem.objective.sense.maximize}[
                direction])

    def _get_primal_values(self):
        try:
            primal_values = self.problem.solution.get_values()
        except CplexSolverError as err:
            raise SolverError(str(err))
        return primal_values

    def _get_reduced_costs(self):
        if self.is_integer:
            raise ValueError("Dual values are not well-defined for integer problems")
        try:
            return self.problem.solution.get_reduced_costs()
        except CplexSolverError as err:
            raise SolverError(str(err))

    def _get_constraint_values(self):
        try:
            return self.problem.solution.get_activity_levels()
        except CplexSolverError as err:
            raise SolverError(str(err))

    def _get_shadow_prices(self):
        if self.is_integer:
            raise ValueError("Dual values are not well-defined for integer problems")
        try:
            return self.problem.solution.get_dual_values()
        except CplexSolverError as err:
            raise SolverError(str(err))

    @property
    def is_integer(self):
        return (self.problem.variables.get_num_integer() + self.problem.variables.get_num_binary()) > 0

    def to_lp(self):
        self.update()
        with TemporaryFilename(suffix=".lp") as tmp_file_name:
            self.problem.write(tmp_file_name)
            with open(tmp_file_name) as tmp_file:
                lp_form = tmp_file.read()
        return lp_form

    def _optimize(self):
        try:
            self.problem.solve()
        except CplexSolverError as err:
            raise SolverError(str(err))
        cplex_status = self.problem.solution.get_status()
        self._original_status = self.problem.solution.get_status_string()
        status = _CPLEX_STATUS_TO_STATUS[cplex_status]
        return status

    def _set_variable_bounds_on_problem(self, var_lb, var_ub):
        lb = [
            (var.name, -cplex.infinity) if val is None else (var.name, float(val)) for var, val in var_lb
        ]
        if len(lb) > 0:
            self.problem.variables.set_lower_bounds(lb)
        ub = [
            (var.name, cplex.infinity) if val is None else (var.name, float(val)) for var, val in var_ub
        ]
        if len(ub) > 0:
            self.problem.variables.set_upper_bounds(ub)

    def _add_variables(self, variables):
        super(Model, self)._add_variables(variables)
        lb, ub, vtype, names, coeff = list(), list(), list(), list(), list()
        for variable in variables:
            if variable.lb is None:
                lb.append(-cplex.infinity)
            else:
                lb.append(variable.lb)
            if variable.ub is None:
                ub.append(cplex.infinity)
            else:
                ub.append(variable.ub)
            vtype.append(_VTYPE_TO_CPLEX_VTYPE[variable.type])
            names.append(variable.name)
            coeff.append(0.)
            variable.problem = self

        if set(vtype) == set(['C']):  # this is needed because CPLEX will automatically set the problem_type to MILP if types are specified
            self.problem.variables.add(coeff, lb=lb, ub=ub, names=names)
        else:
            self.problem.variables.add(coeff, lb=lb, ub=ub, types=vtype, names=names)

    def _remove_variables(self, variables):
        # Not calling parent method to avoid expensive variable removal from sympy expressions
        if self.objective is not None:
            self.objective._expression = self.objective.expression.xreplace({var: 0 for var in variables})
        self.problem.variables.delete([variable.name for variable in variables])
        for variable in variables:
            del self._variables_to_constraints_mapping[variable.name]
            variable.problem = None
            del self._variables[variable.name]
        if not self.is_integer:
            cplex_type = self.problem.get_problem_type()
            if cplex_type in _CPLEX_MIP_TYPES_TO_CONTINUOUS:
                self.problem.set_problem_type(_CPLEX_MIP_TYPES_TO_CONTINUOUS[cplex_type])

    def _add_constraints(self, constraints, sloppy=False):
        super(Model, self)._add_constraints(constraints, sloppy=sloppy)

        linear_constraints = dict(lin_expr=[], senses=[], rhs=[], range_values=[], names=[])
        for constraint in constraints:
            constraint._problem = None  # This needs to be done in order to not trigger constraint._get_expression()
            if constraint.is_Linear:
                offset, coeff_dict, _ = parse_optimization_expression(constraint)

                sense, rhs, range_value = _constraint_lb_and_ub_to_cplex_sense_rhs_and_range_value(
                    constraint.lb,
                    constraint.ub
                )
                indices = [var.name for var in coeff_dict]
                values = [float(val) for val in coeff_dict.values()]
                if constraint.indicator_variable is None:
                    linear_constraints['lin_expr'].append(cplex.SparsePair(ind=indices, val=values))
                    linear_constraints['senses'].append(sense)
                    linear_constraints['rhs'].append(rhs)
                    linear_constraints['range_values'].append(range_value)
                    linear_constraints['names'].append(constraint.name)
                else:
                    if sense == 'R':
                        raise ValueError(
                            'CPLEX does not support indicator constraints that have both an upper and lower bound.')
                    else:
                        # Indicator constraints cannot be added in batch
                        self.problem.indicator_constraints.add(
                            lin_expr=cplex.SparsePair(ind=indices, val=values), sense=sense, rhs=rhs,
                            name=constraint.name,
                            indvar=constraint.indicator_variable.name, complemented=abs(constraint.active_when) - 1)

            elif constraint.is_Quadratic:
                raise NotImplementedError('Quadratic constraints (like %s) are not supported yet.' % constraint)
            else:
                raise ValueError(
                    "CPLEX only supports linear or quadratic constraints. %s is neither linear nor quadratic." % constraint)
            constraint.problem = self
        self.problem.linear_constraints.add(**linear_constraints)

    def _remove_constraints(self, constraints):
        super(Model, self)._remove_constraints(constraints)
        for constraint in constraints:
            if constraint.indicator_variable is not None:
                self.problem.indicator_constraints.delete(constraint.name)
            elif constraint.is_Linear:
                self.problem.linear_constraints.delete(constraint.name)
            elif constraint.is_Quadratic:
                self.problem.quadratic_constraints.delete(constraint.name)

    def _get_quadratic_expression(self, quadratic=None):
        if quadratic is None:
            try:
                quadratic = self.problem.objective.get_quadratic()
            except IndexError:
                return Zero
        terms = []
        for i, sparse_pair in enumerate(quadratic):
            for j, val in zip(sparse_pair.ind, sparse_pair.val):
                i_name, j_name = self.problem.variables.get_names([i, j])
                if i < j:
                    terms.append(val * self._variables[i_name] * self._variables[j_name])
                elif i == j:
                    terms.append(0.5 * val * self._variables[i_name] ** 2)
                else:
                    pass  # Only look at upper triangle
        return add(terms)

    def _get_variable_indices(self, names):
        # Cplex does not keep an index of variable names
        # Getting indices thus takes roughly quadratic time
        # If many indices are required an alternate and faster method is used, where the full name list must only
        # be traversed once
        if len(names) < 1000:
            return self.problem.variables.get_indices(names)
        else:
            name_set = set(names)
            all_names = self.problem.variables.get_names()
            indices = {n: i for i, n in enumerate(all_names) if n in name_set}
            return [indices[n] for n in names]



# from optlang.tests.test_cplex_interface import *
#
# def test_all():
#     try:
#         from nose import run, runmodule, main
#     except ImportError:
#         print('Please install nosetests in order to run optlang tests.')
#     else:
#         # run(argv=[os.path.dirname(os.path.realpath(__file__)), 'test_glpk_interface'])
#         print(os.path.dirname(os.path.abspath(__file__)))
#         tests_dir = os.path.dirname(os.path.abspath(__file__))
#         # import glob
#         # glob.glob()
#         # run(argv=[tests_dir, '-vv'])
#         run(argv=['/Users/niso/Dev/optlang/optlang', '-vv'])



# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Interface to the inspyred heuristic optimization framework

Wraps the GLPK solver by subclassing and extending :class:`Model`,
:class:`Variable`, and :class:`Constraint` from :mod:`interface`.
"""

import logging
import random
import types

log = logging.getLogger(__name__)
import sympy
import interface


class Variable(interface.Variable):
    def __init__(self, *args, **kwargs):
        super(Variable, self).__init__(*args, **kwargs)


class Objective(interface.Objective):
    """docstring for Objective"""

    def __init__(self, expression, *args, **kwargs):
        super(Objective, self).__init__(expression, *args, **kwargs)

    @property
    def value(self):
        return self._value

    def __str__(self):
        if isinstance(self.expression, sympy.Basic):
            return super(Objective, self).__str__()
        else:
            return self.expression.__str__()
            # return ' '.join((self.direction, str(self.expression)))

    @property
    def expression(self):
        return self._expression

    @expression.setter
    def expression(self, value):
        self._expression = value


class VariableBounder(object):
    """This class defines a inspyred like Bounder.__init__.py

    TODO: Make this work also for integer and binary type variables?
    """

    def __init__(self, model):
        self.model = model

    def __call__(self, candidate, args):
        variables = self.model.variables
        bounded_candidate = list()
        for c, variable in zip(candidate, variables):
            if variable.type == 'continuous':
                bounded_candidate.append(max(min(c, variable.ub), variable.lb))
            elif variable.type == 'integer':
                bounded_candidate.append(min(range(variable.lb, variable.ub + 1), key=lambda x: abs(x - c)))
            elif variable.type == 'binary':
                # print min([0, 1], key=lambda x: abs(x-c))
                bounded_candidate.append(min([0, 1], key=lambda x: abs(x - c)))
        return bounded_candidate


class Configuration(interface.EvolutionaryOptimizationConfiguration):
    """docstring for Configuration"""

    class SubConfiguration(object):
        pass

    def __init__(self, *args, **kwargs):
        super(Configuration, self).__init__(*args, **kwargs)
        self._algorithm = inspyred.ec.GA
        self._algorithm.terminator = [inspyred.ec.terminators.time_termination,
                                      inspyred.ec.terminators.generation_termination,
                                      inspyred.ec.terminators.evaluation_termination,
                                      inspyred.ec.terminators.diversity_termination,
                                      inspyred.ec.terminators.average_fitness_termination]
        self.pop_size = 100
        self.seeds = []
        self.max_generations = 1
        self.max_evaluations = None
        self.max_time = None

        self.selector_config = self.SubConfiguration()
        self.selector_config.num_selected = None
        self.selector_config.tournament_size = 2
        self.selector_config.num_elites = 0

        self.variator_config = self.SubConfiguration()
        self.variator_config.mutation_rate = .1
        self.variator_config.crossover_rate = 1.
        self.variator_config.num_crossover_points = 1

        self.topology_config = self.SubConfiguration()
        self.topology_config.neighborhood_size = 5

        self.swarm_config = self.SubConfiguration()
        self.swarm_config.inertia = 0.5
        self.swarm_config.cognitive_rate = 2.1
        self.swarm_config.social_rate = 2.1

    @property
    def selector(self):
        return self._algorithm.selector

    @selector.setter
    def selector(self, value):
        self.algorithm.selector = value

    @property
    def variator(self):
        return self._algorithm.variator

    @variator.setter
    def variator(self, value):
        self._algorithm.variator = value

    @property
    def replacer(self):
        return self._algorithm.replacer

    @replacer.setter
    def replacer(self, value):
        self._algorithm.replacer = value

    @property
    def migrator(self):
        return self._algorithm.migrator

    @migrator.setter
    def migrator(self, value):
        self._algorithm.migrator = value

    @property
    def archiver(self):
        return self._algorithm.archiver

    @archiver.setter
    def archiver(self, value):
        self._algorithm.archiver = value

    @property
    def observer(self):
        return self._algorithm.observer

    @observer.setter
    def observer(self, value):
        self._algorithm.observer = value

    @property
    def terminator(self):
        return self._algorithm.terminator

    @terminator.setter
    def terminator(self, value):
        self._algorithm.terminator = value

    @property
    def topology(self):
        return self._algorithm.topology

    @topology.setter
    def topology(self, value):
        if value == 'Ring':
            self._algorithm.topology = inspyred.swarm.topologies.ring_topology
        elif value == 'Star':
            self._algorithm.topology = inspyred.swarm.topologies.star_topology
        elif isinstance(value, types.FunctionType):
            self._algorithm.topology = value
        else:
            raise ValueError("%s is not a supported topology. Try 'Star' or 'Ring' instead.")

    @property
    def algorithm(self):
        return self._algorithm

    @algorithm.setter
    def algorithm(self, value):
        init = False
        try:
            previous_selector = self._algorithm.selector
            previous_variator = self._algorithm.variator
            previous_replacer = self._algorithm.replacer
            # previous_migrator = self._algorithm.migrator
            # previous_archiver = self._algorithm.archiver
            # previous_observer = self._algorithm.observer
            # previous_terminator = self._algorithm.terminator
        except AttributeError:
            init = True

        if value == "EvolutionaryComputation":
            self._algorithm = inspyred.ec.EvolutionaryComputation
        elif value == "GeneticAlgorithm" or value == "GA":
            self._algorithm = inspyred.ec.GA(random)
        elif value == "ParticleSwarmOptimization" or value == "PSO":
            self._algorithm = inspyred.swarm.PSO(random)
        elif value == "AntColonySystem" or value == "ACS":
            self._algorithm = inspyred.swarm.ACS(random)
        elif value == "EvolutionaryStrategy" or value == "ES":
            self._algorithm = inspyred.ec.ES(random)
        elif value == "DifferentialEvolutionaryAlgorithm" or value == "DEA":
            self._algorithm = inspyred.ec.DEA(random)
        elif value == "SimulatedAnnealing" or value == "SA":
            self._algorithm = inspyred.ec.SA(random)
        elif value == "NSGA2":
            self._algorithm = inspyred.emo.NSGA2(random)
        elif value == "PAES":
            self._algorithm = inspyred.emo.PAES(random)
        elif value == "Pareto":
            self._algorithm = inspyred.emo.Pareto(random)
        else:
            raise ValueError(
                "%s is not a supported. Try one of the following instead:"
                "'GeneticAlgorithm', 'ParticleSwarmOptimization', 'EvolutionaryStrategy'."
                "TODO: be more specific here")
        # self._algorithm.terminator = self._default_terminator
        if init is False:
            self._algorithm.selector = previous_selector
            self._algorithm.variator = previous_variator
            self._algorithm.replacer = previous_replacer
            # previous_migrator = self._algorithm.migrator
            # previous_archiver = self._algorithm.archiver
            # previous_observer = self._algorithm.observer
            # previous_terminator = self._algorithm.terminator
            # TODO: setting a new algorithm should recycle old variators, selectors etc.

    def _evolve_kwargs(self):
        """Filter None keyword arguments. Intended to be passed on to algorithm.evolve(...)"""
        valid_evolve_kwargs = (
            'max_generations', 'max_evaluations', 'pop_size', 'neighborhood_size', 'tournament_size', 'mutation_rate')
        filtered_evolve_kwargs = dict()
        for key in valid_evolve_kwargs:
            attr_value = getattr(self, key)
            if attr_value is not None:
                filtered_evolve_kwargs[key] = attr_value
        # return filtered_evolve_kwargs
        return {}


class Model(interface.Model):
    """Interface"""

    def __init__(self, algorithm=None, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)
        self.configuration = Configuration()
        if algorithm is None:
            self.configuration.algorithm = "GA"
        else:
            self.configuration.algorithm = algorithm
        self._bounder = VariableBounder(self)
        self._generator = self._generator

    def _generator(self, random, args):
        individual = list()
        for variable in self.variables:
            if variable.type == 'continuous':
                individual.append(random.uniform(variable.lb, variable.ub))
            else:
                individual.append(random.choice(range(variable.lb, variable.ub + 1)))
        return individual

    def _evaluator(self, candidates, args):
        fitness = list()
        for candidate in candidates:
            substitution_dict = dict(zip(self.variables, candidate))
            if isinstance(self.objective.expression, sympy.Basic):
                fitness.append(self.objective.expression.subs(substitution_dict))
            else:
                fitness.append(self.objective.expression(substitution_dict))
        return fitness

    # @inspyred.ec.evaluators.evaluator
    # def _evaluate(self, candidate, args):
    #     substitution_dict = dict(zip(self.variables, candidate))
    #     try:
    #         fitness = self.objective.expression.subs(substitution_dict)
    #     except AttributeError:
    #         fitness = self.objective.expression(substitution_dict)
    #     return fitness

    def optimize(self, *args, **kwargs):
        # import pdb; pdb.set_trace();
        final_population = self.configuration.algorithm.evolve(
            generator=self._generator,
            evaluator=self._evaluator,
            bounder=self._bounder,
            pop_size=self.configuration.pop_size,
            maximize={'max': True, 'min': False}[self.objective.direction],
            max_generations=self.configuration.max_generations,
            max_evaluations=self.configuration.max_evaluations,
            neighborhood_size=self.configuration.topology_config.neighborhood_size,
            mutation_rate=self.configuration.variator_config.mutation_rate,
            tournament_size=self.configuration.selector_config.tournament_size
        )
        return final_population


if __name__ == '__main__':
    # from optlang.interface import Objective, Variable
    import numpy
    import inspyred

    x = Variable('x', lb=0, ub=2)
    y = Variable('y', lb=0, ub=2)
    rosenbrock_obj = Objective((1 - x) ** 2 + 100 * (y - x ** 2) ** 2, name="Rosenbrock function", direction='min')
    print("The rosenbrock function:", rosenbrock_obj)
    print("The global minimum at (x,y) = (1,1) is", rosenbrock_obj.expression.subs({x: 1, y: 1}))

    problem = Model(name='rosenbrock', algorithm='PSO')
    # problem = Model(name='rosenbrock')

    problem.objective = rosenbrock_obj


    def my_observer(population, num_generations, num_evaluations, args):
        best = max(population)
        print(('{0:6} -- {1} : {2}'.format(num_generations,
                                           best.fitness,
                                           str(best.candidate))))


    problem.configuration.max_generations = 100
    problem.configuration.terminator = inspyred.ec.terminators.generation_termination
    problem.configuration.observer = my_observer
    problem.configuration.selector = inspyred.ec.selectors.tournament_selection
    final_pop = problem.optimize()
    fitnesses = [individual.fitness for individual in final_pop]
    print(fitnesses)
    print("mean", numpy.mean(fitnesses))
    print("max", numpy.max(fitnesses))
    print("min", numpy.min(fitnesses))
    # print numpy.std(fitnesses)



# Copyright 2016 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from optlang.symbolics import Integer

one = Integer(1)


def parse_optimization_expression(obj, linear=True, quadratic=False, expression=None, **kwargs):
    """
    Function for parsing the expression of a Constraint or Objective object.

    Parameters
    ----------
    object: Constraint or Objective
        The optimization expression to be parsed
    linear: Boolean
        If True the expression will be assumed to be linear
    quadratic: Boolean
        If True the expression will be assumed to be quadratic
    expression: Sympy expression or None (optional)
        An expression can be passed explicitly to avoid getting the expression from the solver.
        If this is used then 'linear' or 'quadratic' should be True.

    If both linear and quadratic are False, the is_Linear and is_Quadratic methods will be used to determine how it should be parsed

    Returns
    ----------
    A tuple of (linear_coefficients, quadratic_coefficients)
        linear_coefficients is a dictionary of {variable: coefficient} pairs
        quadratic_coefficients is a dictionary of {frozenset(variables): coefficient} pairs
    """
    if expression is None:
        expression = obj.expression

    if not (linear or quadratic):
        if obj.is_Linear:
            linear = True
        elif obj.is_Quadratic:
            quadratic = True
        else:
            raise ValueError("Expression is not linear or quadratic. Other expressions are not currently supported.")

    assert linear or quadratic

    if quadratic:
        offset, linear_coefficients, quadratic_coefficients = _parse_quadratic_expression(expression, **kwargs)
    else:
        offset, linear_coefficients = _parse_linear_expression(expression, **kwargs)
        quadratic_coefficients = {}

    return offset, linear_coefficients, quadratic_coefficients


def _parse_linear_expression(expression, expanded=False, **kwargs):
    """
    Parse the coefficients of a linear expression (linearity is assumed).

    Returns a dictionary of variable: coefficient pairs.
    """
    offset = 0
    constant = None
    expression = expression.expand()

    if expression.is_Add:
        coefficients = expression.as_coefficients_dict()
    elif expression.is_Mul:
        coefficients = {expression.args[1]: expression.args[0]}
    elif expression.is_Symbol:
        coefficients = {expression: 1}
    elif expression.is_Number:
        return float(expression), {}
    else:
        raise ValueError("Expression {} seems to be invalid".format(expression))

    for var in coefficients:
        if not (var.is_Symbol):
            if var == one:
                constant = var
                offset = float(coefficients[var])
            elif expanded:
                raise ValueError("Expression {} seems to be invalid".format(expression))
            else:
                coefficients = _parse_linear_expression(expression, expanded=True, **kwargs)
    if constant is not None:
        del coefficients[constant]
    return offset, coefficients


def _parse_quadratic_expression(expression, expanded=False):
    """
    Parse a quadratic expression. It is assumed that the expression is known to be quadratic or linear.

    The 'expanded' parameter tells whether the expression has already been expanded. If it hasn't the parsing
    might fail and will expand the expression and try again.
    """
    linear_coefficients = {}
    quadratic_coefficients = {}
    offset = 0
    expression = expression.expand()

    if expression.is_Number:  # Constant expression, no coefficients
        return float(expression), linear_coefficients, quadratic_coefficients

    if expression.is_Mul:
        terms = (expression,)
    elif expression.is_Add:
        terms = expression.args
    else:
        raise ValueError("Expression of type {} could not be parsed.".format(type(expression)))

    try:
        for term in terms:
            if term.is_Number:
                offset += float(term)
                continue
            if term.is_Pow:
                term = 1.0 * term
            assert term.is_Mul, "What is this? {}".format(type(term))
            factors = term.args
            coef = factors[0]
            vars = factors[1:]
            assert len(vars) <= 2, "This should not happen. Is this expression quadratic?"
            if len(vars) == 2:
                key = frozenset(vars)
                quadratic_coefficients[key] = quadratic_coefficients.get(key, 0) + coef
            else:
                var = vars[0]
                if var.is_Symbol:
                    linear_coefficients[var] = linear_coefficients.get(var, 0) + coef
                elif var.is_Pow:
                    var, exponent = var.args
                    if exponent != 2:
                        raise ValueError("The expression is not quadratic")
                    key = frozenset((var,))
                    quadratic_coefficients[key] = quadratic_coefficients.get(key, 0) + coef
        if quadratic_coefficients:
            assert all(var.is_Symbol for var in frozenset.union(*quadratic_coefficients))  # Raise an exception to trigger expand
        if linear_coefficients:
            assert all(var.is_Symbol for var in linear_coefficients)
    except Exception as e:
        if expanded:
            raise e
        else:
            # Try to expand the expression and parse it again
            return _parse_quadratic_expression(expression.expand(), expanded=True)

    return offset, linear_coefficients, quadratic_coefficients
